from __future__ import annotations

"""
Temporal Workflow Mixins: Activities
"""

import typing as t
from .base import BaseTemporalMixin, MixinKinds

class TemporalActivityMixin(BaseTemporalMixin):
    """
    [Temporal] Activity Mixin that will automatically registered
    """

    mixin_kind: t.Optional[MixinKinds] = 'activity'
    _is_subclass_: t.Optional[bool] = True

    # These will be passed to the activity defn decorator
    no_thread_cancel_exception: t.Optional[bool] = None
    dynamic: t.Optional[bool] = None

    include_funcs: t.Optional[t.List[str]] = None
    exclude_funcs: t.Optional[t.List[str]] = None

    @classmethod
    def configure_registered(cls, **kwargs):
        """
        Configures the registered activity
        """
        _kws = {k: kwargs.pop(k) for k in kwargs if k in {'no_thread_cancel_exception', 'dynamic', 'include_funcs', 'exclude_funcs'}}
        new = super().configure_registered(**kwargs)
        if 'no_thread_cancel_exception' in _kws: new.no_thread_cancel_exception = _kws.pop('no_thread_cancel_exception')
        if 'dynamic' in _kws: new.dynamic = _kws.pop('dynamic')
        if 'include_funcs' in _kws: new.include_funcs = _kws.pop('include_funcs')
        if 'exclude_funcs' in _kws: new.exclude_funcs = _kws.pop('exclude_funcs')
        return new

    @classmethod
    def _on_register_hook_(cls):
        """
        Runs the activity hook
        """
        pass

    def _on_init_hook_(self):
        """
        Runs the activity init hook
        """
        pass




from __future__ import annotations

import json
import enum
import time
import asyncio
import functools
import contextlib
from lzl.api import aiohttpx
from typing import Optional, Type, Any, Union, List, Dict, Iterator, TypeVar, AsyncIterator, Generator, AsyncGenerator, overload, TYPE_CHECKING

from lzo.types import (
    Literal, 
    eproperty, 
    BaseModel,
    validator, 
    root_validator, 
    Field,
    PYDANTIC_VERSION, 
)
from lzl.types.base import get_pydantic_schema, field_validator
from lzl.api.openai.version import DEFAULT_CHAT_MODEL
from lzl.api.openai.types.handlers import ModelContextHandler
from lzl.api.openai.types.base import BaseResource, Usage
from lzl.api.openai.types.responses import BaseResponse
from lzl.api.openai.types.routes import BaseRoute
from lzl.api.openai.utils import logger, parse_stream, aparse_stream, resolve_json
from lzl.api.openai.utils.schemas import generate_json_schema, function_generate_json_schema

__all__ = [
    'ChatMessage',
    'ChatChoice',
    'Function',
    'ChatObject',
    'CompletionResponse',
    'CompletionRoute',
]

SchemaObj = TypeVar("SchemaObj", bound=BaseModel)
SchemaType = TypeVar("SchemaType", bound=Type[BaseModel])




class MessageKind(str, enum.Enum):
    CONTENT = 'content'
    ROLE = 'role'
    FUNCTION_CALL = 'function_call'

    @classmethod
    def from_choice(cls, choice: Dict[str, Any]) -> 'MessageKind':
        """
        Returns the message kind from the choice
        """
        if 'role' in choice['delta']:
            return cls.ROLE
        elif 'content' in choice['delta']:
            return cls.CONTENT
        elif 'function_call' in choice['delta']:
            return cls.FUNCTION_CALL
        raise ValueError(f'Invalid choice: {choice}')



class StreamedChatMessage(BaseResource):
    kind: MessageKind
    value: Union[Dict[str, Any], str]

class FunctionCall(BaseResource):
    name: str
    arguments: Optional[Union[str, Dict[str, Any]]] = None

    @field_validator('arguments')
    def validate_arguments(cls, v) -> Dict[str, Any]:
        """
        Try to load the arguments as json
        """
        if isinstance(v, dict):
            return v
        elif isinstance(v, str):
            with contextlib.suppress(Exception):
                return json.loads(v)
        return v

class ToolCall(BaseResource):
    id: Optional[str] = None
    type: Optional[str] = None
    function: Optional[FunctionCall] = None


# TODO Add support for name
class ChatMessage(BaseResource):
    content: Optional[str] = None
    role: Optional[str] = "user"
    function_call: Optional[FunctionCall] = None
    tool_calls: Optional[List[ToolCall]] = None
    name: Optional[str] = None

    def dict(self, *args, exclude_none: bool = True, **kwargs):
        return self.model_dump(*args, exclude_none = exclude_none, **kwargs)


class ChatChoice(BaseResource):
    message: ChatMessage
    index: int
    logprobs: Optional[Any] = None
    finish_reason: Optional[str] = None

    def __getitem__(self, key: str) -> Any:
        """
        Mimic dict
        """
        return getattr(self, key)

class Function(BaseResource):
    """
    Represents a function
    """
    # Must be a-z, A-Z, 0-9, or contain underscores and dashes
    if PYDANTIC_VERSION == 2:
        name: str = Field(..., max_length = 64, pattern = r'^[a-zA-Z0-9_]+$')
    else:
        name: str = Field(..., max_length = 64, regex = r'^[a-zA-Z0-9_]+$')
    
    parameters: Union[Dict[str, Any], SchemaType, str]
    description: Optional[str] = None
    # Add optional enforcement of structured outputs
    # strict: Optional[bool] = None

    source_object: Optional[Union[SchemaType, Any]] = Field(default = None, exclude = True)

    @root_validator(pre = True)
    def validate_parameters(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validate the parameters
        """
        if params := values.get('parameters'):
            if isinstance(params, dict):
                pass
            elif issubclass(params, BaseModel) or isinstance(params, type(BaseModel)):
                values['parameters'] = get_pydantic_schema(params)
                # params.schema()
                values['source_object'] = params
            elif isinstance(params, str):
                try:
                    values['parameters'] = json.loads(params)
                except Exception as e:
                    raise ValueError(f'Invalid JSON: {params}, {e}. Must be a dict or pydantic BaseModel.') from e
            else:
                # logger.warning(f'Invalid parameters: {params}. Must be a dict or pydantic BaseModel.')
                raise ValueError(f'Parameters must be a dict or pydantic BaseModel. Provided: {type(params)}')
        return values

class JSONSchema(BaseResource):
    """
    Represents a JSON Schema
    """
    if PYDANTIC_VERSION == 2:
        name: str = Field(..., max_length = 64, pattern = r'^[a-zA-Z0-9_]+$')
    else:
        name: str = Field(..., max_length = 64, regex = r'^[a-zA-Z0-9_]+$')
    
    # This raises a warning Field name "schema" in "JSONSchema" shadows an attribute in parent "BaseModel"
    # So we alias it to jschema
    jschema: Union[Dict[str, Any], SchemaType, str] = Field(..., alias = 'schema')
    description: Optional[str] = None
    strict: Optional[bool] = True
    from_function: Optional[bool] = Field(None, exclude = True)
    source_object: Optional[Union[SchemaType, Any]] = Field(default = None, exclude = True)

    @root_validator(pre = True)
    def validate_json_schema(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validate the schema
        """
        from_func = values.get('from_function', False)
        if schema := values.get('schema'):
            if isinstance(schema, dict):
                pass
            elif issubclass(schema, BaseModel) or isinstance(schema, type(BaseModel)):
                values['schema'] = function_generate_json_schema(schema) if from_func else generate_json_schema(schema)
                values['source_object'] = schema
            elif isinstance(schema, str):
                try:
                    values['schema'] = json.loads(schema)
                except Exception as e:
                    raise ValueError(f'Invalid JSON: {schema}, {e}. Must be a dict or pydantic BaseModel.') from e
            else:
                # logger.warning(f'Invalid parameters: {params}. Must be a dict or pydantic BaseModel.')
                raise ValueError(f'Schema must be a dict or pydantic BaseModel. Provided: {type(schema)}')
        return values


class Tool(BaseResource):
    """
    Represents a tool 
    """
    type: Optional[str] = 'function'
    function: Optional[Function] = None


class ChatObject(BaseResource):
    messages: Union[List[ChatMessage], str]
    model: Optional[str] = DEFAULT_CHAT_MODEL

    max_tokens: Optional[int] = None
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    stream: Optional[bool] = False
    logprobs: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    logit_bias: Optional[Dict[str, float]] = None
    user: Optional[str] = None

    functions: Optional[List[Function]] = None
    function_call: Optional[Union[str, Dict[str, str]]] = None

    # v2 Params
    # response_format: Optional[Dict[str, Literal['json_object', 'text']]] = None
    # Don't enforce response_format
    json_schema: Optional[Union[JSONSchema, Dict[str, Any]]] = None # Field(default = None, exclude = True)
    response_format: Optional[Dict[str, Any]] = None
    seed: Optional[int] = None

    # tools: Optional[Union[List[Function], List[Dict[str, Union[str, Function]]]]] = None
    tools: Optional[List[Tool]] = None
    tool_choice: Optional[Union[str, Union[str, Dict[str, str]]]] = None

    # Extra Params
    validate_max_tokens: Optional[bool] = Field(default = False, exclude = True)
    validate_model_aliases: Optional[bool] = Field(default = False, exclude = True)
    # api_version: Optional[str] = None

    # @validator('messages', pre = True, always = True)
    @field_validator('messages', mode = 'before')
    def validate_messages(cls, v) -> List[ChatMessage]:
        """
        Validate the Input Messages
        """
        vals = []
        if not isinstance(v, list):
            v = [v]
        for i in v:
            if isinstance(i, dict):
                vals.append(ChatMessage.model_validate(i))
            elif isinstance(i, str):
                vals.append(ChatMessage(content = i))
            else:
                vals.append(i)
        return vals

    # @validator('model', pre=True, always=True)
    @field_validator('model', mode = 'before')
    def validate_model(cls, v, values: Dict[str, Any]) -> str:
        """
        Validate the model
        """
        if not v:
            if values.get('engine'):
                v = values.get('engine')
            elif values.get('deployment'):
                v = values.get('deployment')
        with contextlib.suppress(KeyError):
            v = ModelContextHandler.resolve_model_name(v)
        # if values.get('validate_model_aliases', False):
        #     v = ModelContextHandler[v].name
        return v
        
    
    @validator('temperature')
    def validate_temperature(cls, v: float) -> float:
        """
        Min Temperature is 0.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-temperature
        """
        return None if v is None else max(0.0, v)
    
    @validator('top_p')
    def validate_top_p(cls, v: float) -> float:
        """
        Min Top Probability is 0.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-top-probability
        """
        return None if v is None else max(0.0, v)

    @validator('logprobs')
    def validate_logprobs(cls, v: int) -> int:
        """
        Max logprobs is 5
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-logprobs
        """
        return None if v is None else max(0, min(v, 5))
    
    @validator('presence_penalty')
    def validate_presence_penalty(cls, v: float) -> float:
        """
        Min Presence Penalty is -2.0, Max is 2.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-presence-penalty
        """
        return None if v is None else max(0.0, min(v, 2.0))
    
    @validator('frequency_penalty')
    def validate_frequency_penalty(cls, v: float) -> float:
        """
        Min Frequency Penalty is -2.0, Max is 2.0
        https://beta.openai.com/docs/api-reference/completions/create#completions/create-frequency-penalty
        """
        return None if v is None else max(0.0, min(v, 2.0))

    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        return super().dict(*args, exclude = exclude, **kwargs)

    @root_validator(pre = True)
    def validate_obj(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validate the object
        """

        is_azure = values.pop('is_azure', False)
        if values.get('functions'):
            if not all(isinstance(f, Function) for f in values['functions']):
                values['functions'] = [Function(**f) for f in values['functions']]
            
            if not values.get('function_call'):
                values['function_call'] = 'auto'
        
        if values.get('json_schema'):
            if not isinstance(values['json_schema'], JSONSchema):
                values['json_schema'] = JSONSchema(**values['json_schema'])
            values['response_format'] = {
                'type': 'json_schema',
                'json_schema': values['json_schema'],
            }

        if values.get('tools'):
            tools = []
            for tool in values['tools']:
                if isinstance(tool, Tool):
                    tools.append(tool)
                elif isinstance(tool, dict):
                    # This should be the correct format
                    if tool.get('function'):
                        tools.append(Tool(**tool))
                    else:
                        # This is previously supported format
                        tools.append(Tool(function = Function(**tool)))
                else:
                    raise ValueError(f'Invalid tool: {tool}')
            values['tools'] = tools
            if not values.get('tool_choice'):
                values['tool_choice'] = 'auto'

            # Auto set to json_object if tools are present
            # if values.get('response_format') is None:
            #     values['response_format'] = {'type': 'json_object'}

        # Validate for Azure
        if is_azure and values.get('tools') and not values.get('functions'):
            # Convert tools to functions
            values['functions'] = [tool.function for tool in values['tools']]
            if not values.get('function_call'):
                values['function_call'] = 'auto'

        return values


class ChatResponse(BaseResponse):
    choices: Optional[List[ChatChoice]] = None
    choice_model: Optional[Type[BaseResource]] = ChatChoice
    input_object: Optional[ChatObject] = None

    @eproperty
    def messages(self) -> List[ChatMessage]:
        """
        Returns the messages for the completions
        """
        if self.choices_results:
            return [choice.message for choice in self.choices]
        return self.response.text
    
    @eproperty
    def json_schema(self) -> Optional[Union[SchemaType, Dict[str, Any]]]:
        """
        Returns the JSON Schema for the completions
        """
        if self.choices_results:
            if self.input_object.json_schema:
                message = self.messages[0]
                if self.input_object.json_schema.source_object:
                    return self.input_object.json_schema.source_object.model_validate_json(message.content)
                if not isinstance(message.content, dict):
                    message.content = json.loads(message.content)
                return message.content
        return None
    

    @eproperty
    def has_json_schema(self) -> bool:
        """
        Returns whether the response has a JSON Schema
        """
        return bool(self.input_object.json_schema)
    
    @eproperty
    def has_tools(self) -> bool:
        """
        Returns whether the response has tools
        """
        return bool(self.input_object.tools)
    
    @eproperty
    def function_results(self) -> List[FunctionCall]:
        """
        Returns the function results for the completions
        """
        return [msg.function_call for msg in self.messages if msg.function_call]
    
    @eproperty
    def function_result_objects(self) -> List[Union[SchemaObj, Dict[str, Any]]]:
        """
        Returns the function result objects for the completions
        """
        results = []
        source_function: Function = self.input_object.functions[0] if self.input_object.function_call == "auto" else (
            [
                f for f in self.input_object.functions if f.name == self.input_object.function_call['name']
            ]
        )[0]

        for func_result in self.function_results:
            if source_function.source_object:
                if not isinstance(func_result.arguments, dict):
                    try:
                        func_result.arguments = resolve_json(func_result.arguments)
                    except Exception as e:
                        logger.error('Could not resolve function arguments. Skipping.')
                        continue
                try:
                    results.append(source_function.source_object(**func_result.arguments))
                    continue
                except Exception as e:
                    logger.error(e)
            if isinstance(func_result.arguments, dict):
                results.append(func_result.arguments)
            else:
                try:
                    results.append(resolve_json(func_result.arguments))
                except Exception as e:

                    logger.error(e)
                    results.append(func_result.arguments)

        return results
    
    @eproperty
    def has_functions(self) -> bool:
        """
        Returns whether the response has functions
        """
        return bool(self.input_object.functions)
    

    @eproperty
    def tool_results(self) -> List[FunctionCall]:
        """
        Returns the tool results for the completions
        """
        # if self.has_tools:
        #     return [msg.tool_calls[0].function for msg in self.messages if msg.tool_calls]
        return [msg.function_call for msg in self.messages if msg.function_call]
    
    @eproperty
    def tool_result_objects(self) -> List[Union[SchemaObj, Dict[str, Any]]]:
        """
        Returns the tool result objects for the completions
        """
        results = []
        source_function: Function = self.input_object.tools[0].function.source_object if self.input_object.tool_choice == "auto" else (
            [
                t.function for t in self.input_object.tools if t.function.name == self.input_object.tool_choice['function']['name']
            ]
        )[0]

        for tool_result in self.tool_results:
            if source_function.source_object:
                if not isinstance(tool_result.arguments, dict):
                    try:
                        tool_result.arguments = resolve_json(tool_result.arguments)
                    except Exception as e:
                        logger.error('Could not resolve function arguments. Skipping.')
                        continue
                try:
                    results.append(source_function.source_object(**tool_result.arguments))
                    continue
                except Exception as e:
                    logger.error(e)
            if isinstance(tool_result.arguments, dict):
                results.append(tool_result.arguments)
            else:
                try:
                    results.append(resolve_json(tool_result.arguments))
                except Exception as e:

                    logger.error(e)
                    results.append(tool_result.arguments)

        return results

    @eproperty
    def has_tools(self) -> bool:
        """
        Returns whether the response has tools
        """
        return bool(self.input_object.tools)

    @eproperty
    def input_text(self) -> str:
        """
        Returns the input text for the input prompt
        """
        return '\n'.join([f'{msg.role}: {msg.content}' for msg in self.input_object.messages])

    @eproperty
    def input_messages(self) -> List[ChatMessage]:
        """
        Returns the input messages for the input prompt
        """
        return self.input_object.messages

    @eproperty
    def text(self) -> str:
        """
        Returns the text for the chat response
        """
        if self.choices_results:
            return '\n'.join([f'{msg.role}: {msg.content}' for msg in self.messages])
        return self.response.text


    @eproperty
    def only_text(self) -> str:
        """
        Returns the text for the chat response without the role
        """
        if self.has_tools:
            data = []
            for tool_obj in self.tool_result_objects:
                if isinstance(tool_obj, BaseModel):
                    data.append(tool_obj.model_dump())
                else:
                    data.append(tool_obj)
            return json.dumps(data, indent = 2)
        if self.has_functions:
            data = []
            for func_obj in self.function_result_objects:
                if isinstance(func_obj, BaseModel):
                    data.append(func_obj.model_dump())
                else:
                    data.append(func_obj)
            return json.dumps(data, indent = 2)
        if self.choices_results:
            return '\n'.join([msg.content for msg in self.messages])
        return self.response.text

    @eproperty
    def chat_model(self) -> Optional[str]:
        """
        Returns the model for the completions
        """
        return self.input_object.model or None
        # return OpenAIModel(value=self.model, mode='chat') if self.model else None
    
    @eproperty
    def openai_model(self):
        """
        Returns the model for the completions
        """
        return self.headers.get('openai-model', self.chat_model)
    
    def _validate_usage(self):
        """
        Validate usage
        """
        if self.usage and self.usage.total_tokens and self.usage.prompt_tokens: return
        if self.response.status_code == 200:
            self.usage = Usage(
                prompt_tokens = ModelContextHandler.count_chat_tokens(self.input_messages, model_name = self.openai_model),
                completion_tokens = ModelContextHandler.count_chat_tokens(self.messages, model_name = self.openai_model),
            )
            self.usage.total_tokens = self.usage.prompt_tokens + self.usage.completion_tokens


    @eproperty
    def consumption(self) -> int:
        """
        Returns the consumption for the completions
        """ 
        self._validate_usage()
        return ModelContextHandler.get_consumption_cost(
            model_name = self.openai_model,
            usage = self.usage,
        )
    

    def dict(self, *args, exclude: Any = None, **kwargs):
        """
        Returns the dict representation of the response
        """
        return super().dict(*args, exclude = exclude, **kwargs)
    

    def __getitem__(self, key: str) -> Any:
        """
        Mimic dict
        """
        return getattr(self, key)
    

    def parse_stream_item(self, item: Union[Dict, Any], **kwargs) -> Optional[StreamedChatMessage]:
        """
        Parses a single stream item
        """
        # logger.info(f'Item: {item}')
        if not item['choices']: return None
        choice = item['choices'][0]
        if choice['finish_reason'] in ['stop', 'function_call']:
            return None
        kind = MessageKind.from_choice(choice)
        return StreamedChatMessage(
            kind = kind,
            value = choice['delta'].get(kind.value, '')
        )
    
    def handle_stream(
        self,
        response: aiohttpx.Response,
        streaming: Optional[bool] = False,
    ) -> Iterator[Dict]:  # sourcery skip: low-code-quality
        """
        Handle the stream response
        """
        results: Dict[str, Dict[str, Any]] = {}
        for line in parse_stream(response):
            try:
                item = json.loads(line)
                if streaming:
                    yield item
                self.handle_stream_metadata(item)
                for n, choice in enumerate(item['choices']):
                    if not results.get(n):
                        results[n] = {
                            'index': choice['index'],
                            'message': {
                                'role': choice['delta'].get('role', ''),
                                'content': choice['delta'].get('content', ''),
                            }
                        }
                        if 'function_call' in choice['delta']:
                            results[n]['message']['function_call'] = choice['delta']['function_call']
                            self.usage.completion_tokens += 4
                        
                        # every message follows <im_start>{role/name}\n{content}<im_end>\n
                        self.usage.completion_tokens += 4

                    elif choice['finish_reason'] != 'stop':
                        for k,v in choice['delta'].items():
                            if k == 'function_call' and v:
                                for fck, fcv in v.items():
                                    if not results[n]['message'][k].get(fck):
                                        results[n]['message'][k][fck] = fcv
                                    else:
                                        results[n]['message'][k][fck] += fcv

                            elif v: results[n]['message'][k] += v
                            self.usage.completion_tokens += 1

                    else:
                        results[n]['finish_reason'] = choice['finish_reason']
                        self.usage.completion_tokens += 2  # every reply is primed with <im_start>assistant
                        compl = results.pop(n)
                        if streaming:
                            self.handle_resource_item(item = compl)
                        else:
                            yield compl

            except Exception as e:
                logger.error(f'Error: {line}: {e}')
        self._stream_consumed = True
        for remaining_result in results.values():
            if streaming:
                self.handle_resource_item(item = remaining_result)
            else:
                yield remaining_result
        if not self.usage.prompt_tokens:
            self.usage.prompt_tokens = ModelContextHandler.count_chat_tokens(
                messages = self.input_messages,
                model_name = self.openai_model
            )
        self.usage.total_tokens = self.usage.completion_tokens + self.usage.prompt_tokens
        
    
    async def ahandle_stream(
        self,
        response: aiohttpx.Response,
        streaming: Optional[bool] = False,
    ) -> AsyncIterator[Dict]:  # sourcery skip: low-code-quality
        """
        Handles the streaming response
        """
        results: Dict[str, Dict[str, Any]] = {}
        async for line in aparse_stream(response):
            # logger.info(f'line: {line}')
            try:
                item = json.loads(line)
                self.handle_stream_metadata(item)
                if streaming:
                    yield item

                # logger.info(f'item: {item}')
                for n, choice in enumerate(item['choices']):
                    if not results.get(n):
                        results[n] = {
                            'index': choice['index'],
                            'message': {
                                'role': choice['delta'].get('role', ''),
                                'content': choice['delta'].get('content', ''),
                            }
                        }
                        if 'function_call' in choice['delta']:
                            results[n]['message']['function_call'] = choice['delta']['function_call']
                            self.usage.completion_tokens += 4
                        
                        # every message follows <im_start>{role/name}\n{content}<im_end>\n
                        self.usage.completion_tokens += 4

                    elif choice['finish_reason'] != 'stop':
                        for k,v in choice['delta'].items():
                            if k == 'function_call' and v:
                                for fck, fcv in v.items():
                                    if not results[n]['message'][k].get(fck):
                                        results[n]['message'][k][fck] = fcv
                                    else:
                                        results[n]['message'][k][fck] += fcv
                            elif v: results[n]['message'][k] += v
                            self.usage.completion_tokens += 1

                    else:
                        results[n]['finish_reason'] = choice['finish_reason']
                        self.usage.completion_tokens += 2  # every reply is primed with <im_start>assistant
                        compl = results.pop(n)
                        if streaming:
                            self.handle_resource_item(item = compl)
                        else:
                            yield compl


            except Exception as e:
                logger.trace(f'Error: {line}', e)
        # self.ctx.stream_consumed = True
        self._stream_consumed = True
        for remaining_result in results.values():
            if streaming:
                self.handle_resource_item(item = remaining_result)
            else:
                yield remaining_result
        
        if not self.usage.prompt_tokens:
            self.usage.prompt_tokens = ModelContextHandler.count_chat_tokens(
                messages = self.input_messages,
                model_name = self.openai_model
            )
        self.usage.total_tokens = self.usage.completion_tokens + self.usage.prompt_tokens

    if TYPE_CHECKING:
        def stream(self, **kwargs) -> Generator[StreamedChatMessage, None, None]:
            """
            Streams the response
            """
            ...

        async def astream(self, **kwargs) -> AsyncGenerator[StreamedChatMessage, None]:
            """
            Streams the response
            """
            ...
    


class ChatRoute(BaseRoute):
    input_model: Optional[Type[BaseResource]] = ChatObject
    response_model: Optional[Type[BaseResource]] = ChatResponse
    api_resource: Optional[str] = Field(default = 'chat/completions')
    root_name: Optional[str] = Field(default = 'chat')

    def get_serialization_kwargs(
        self,
        data: 'ChatObject',
    ) -> Dict[str, Any]:
        """
        Returns the serialization kwargs
        """
        opts = {
            'exclude_none': self.exclude_null,
        }
        if data.json_schema: 
            opts['by_alias'] = True
            opts['exclude'] = {'json_schema'}
        return opts


    @overload
    def create(
        self,
        messages: Union[List[ChatMessage], List[Dict[str, str]]],
        model: Optional[str] = None,
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        n: Optional[int] = None,
        stream: Optional[bool] = None,
        logprobs: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[Dict[str, float]] = None,
        user: Optional[str] = None,
    ) -> ChatResponse:
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = OpenAI.chat.create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`
        
        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096 / 8182 / 32,768). If max_tokens is not provided, the model will use the maximum number of tokens
        Default: None

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional): CURRENTLY NOT SUPPORTED
        Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`
        
        Returns a ChatResponse
        """
        ...
    
    @overload
    def create(
        self,
        messages: Union[List[ChatMessage], List[Dict[str, str]]],
        model: Optional[str] = None,
        functions: Optional[List[Function]] = None,
        function_call: Optional[Union[str, Dict[str, str]]] = None,
        response_format: Optional[Dict[str, Literal['json_object', 'text']]] = None,
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        header_cache_keys: Optional[List[str]] = None,
        **kwargs
    ) -> ChatResponse:
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = OpenAI.chat.create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    functions = [{'name': 'test', 'parameters': {'test': 'test'}}],
        >>>    function_call = 'auto'
        >>> )
        ```
        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`

        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.

        :temperature (optional): What sampling temperature to use. Higher values means

        :functions (optional): A list of dictionaries representing the functions to call

        :function_call (optional): The name of the function to call

        :auto_retry (optional): Whether to automatically retry the request if it fails

        :auto_retry_limit (optional): The maximum number of times to retry the request

        :header_cache_keys (optional): A list of keys to use for caching the headers

        Returns a ChatResponse
        """
        ...
        

    def create(
        self, 
        input_object: Optional[ChatObject] = None,
        parse_stream: Optional[bool] = True,
        timeout: Optional[int] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        header_cache_keys: Optional[List[str]] = None,
        **kwargs
    ) -> ChatResponse:  # sourcery skip: low-code-quality
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = OpenAI.chat.create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`
        
        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096 / 8182 / 32,768). If max_tokens is not provided, the model will use the maximum number of tokens
        Default: None

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional):
        Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.
        Default: `None`

        :functions (optional): A list of dictionaries representing the functions to call
        
        :function_call (optional): The name of the function to call. Default: `auto` if functions are provided

        :response_format (optional): The format of the response. Default: `text`

        :seed (optional): An integer seed for random sampling. Must be between 0 and 2**32 - 1

        :tools (optional): A list of dictionaries representing the tools to use

        :tool_choice (optional): The name of the tool to use. Default: `auto` if tools are provided

        :auto_retry (optional): Whether to automatically retry the request if it fails due to a rate limit error.

        :auto_retry_limit (optional): The maximum number of times to retry the request if it fails due to a rate limit error.

        Returns: `ChatResponse`
        """
        return super().create(
            input_object = input_object,
            parse_stream = parse_stream,
            timeout = timeout,
            auto_retry = auto_retry,
            auto_retry_limit = auto_retry_limit,
            header_cache_keys = header_cache_keys,
            **kwargs
        )


    @overload
    async def async_create(
        self,
        messages: Union[List[ChatMessage], List[Dict[str, str]]],
        model: Optional[str] = None,
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        n: Optional[int] = None,
        stream: Optional[bool] = None,
        logprobs: Optional[int] = None,
        stop: Optional[Union[str, List[str]]] = None,
        presence_penalty: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        logit_bias: Optional[Dict[str, float]] = None,
        user: Optional[str] = None,
    ) -> ChatResponse:
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = await OpenAI.chat.async_create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`
        
        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096 / 8182 / 32,768). If max_tokens is not provided, the model will use the maximum number of tokens
        Default: None

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional):
        Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`
        
        Returns a ChatResponse
        """
        ...
    
    @overload
    async def async_create(
        self,
        messages: Union[List[ChatMessage], List[Dict[str, str]]],
        model: Optional[str] = None,
        functions: Optional[List[Function]] = None,
        function_call: Optional[Union[str, Dict[str, str]]] = None,
        response_format: Optional[Dict[str, Literal['json_object', 'text']]] = None,
        json_schema: Optional[Union[JSONSchema, Dict[str, Any]]] = None,
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        header_cache_keys: Optional[List[str]] = None,
        **kwargs
    ) -> ChatResponse:
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = await OpenAI.chat.async_create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    functions = [{'name': 'test', 'parameters': {'test': 'test'}}],
        >>>    function_call = 'auto'
        >>> )
        ```
        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`

        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.

        :temperature (optional): What sampling temperature to use. Higher values means

        :functions (optional): A list of dictionaries representing the functions to call

        :function_call (optional): The name of the function to call

        :auto_retry (optional): Whether to automatically retry the request if it fails

        :auto_retry_limit (optional): The maximum number of times to retry the request

        :header_cache_keys (optional): A list of keys to use for caching the headers

        Returns a ChatResponse
        """
        ...

    async def async_create(
        self, 
        input_object: Optional[ChatObject] = None,
        parse_stream: Optional[bool] = True,
        timeout: Optional[int] = None,
        auto_retry: Optional[bool] = False,
        auto_retry_limit: Optional[int] = None,
        header_cache_keys: Optional[List[str]] = None,
        **kwargs
    ) -> ChatResponse:  # sourcery skip: low-code-quality
        """
        Creates a chat response for the provided prompt and parameters

        Usage:

        ```python
        >>> result = await OpenAI.chat.async_create(
        >>>    messages = [{'content': 'say this is a test'}],
        >>>    max_tokens = 4,
        >>>    stream = True
        >>> )
        ```

        **Parameters:**

        :model (required): ID of the model to use. You can use the List models API 
        to see all of your available models,  or see our Model overview for descriptions of them.
        Default: `gpt-3.5-turbo`
        
        :messages: The messages to generate chat completions for, in the chat format.

        :max_tokens (optional): The maximum number of tokens to generate in the completion.
        The token count of your prompt plus `max_tokens` cannot exceed the model's context length. 
        Most models have a context length of 2048 tokens (except for the newest models, which 
        support 4096).
        Default: `16`

        :temperature (optional): What sampling temperature to use. Higher values means 
        the model will take more risks. Try 0.9 for more creative applications, and 0 (argmax sampling) 
        for ones with a well-defined answer. We generally recommend altering this or `top_p` but not both.
        Default: `1.0`

        :top_p (optional): An alternative to sampling with `temperature`, called nucleus 
        sampling, where the model considers the results of the tokens with `top_p` probability mass. 
        So `0.1` means only  the tokens comprising the top 10% probability mass are considered.
        We generally recommend altering this or `temperature` but not both
        Default: `1.0`

        :n (optional): How many completions to generate for each prompt.
        Note: Because this parameter generates many completions, it can quickly 
        consume your token quota. Use carefully and ensure that you have reasonable 
        settings for `max_tokens` and stop.
        Default: `1`

        :stream (optional):
        Whether to stream back partial progress. 
        If set, tokens will be sent as data-only server-sent events as they become 
        available, with the stream terminated by a `data: [DONE]` message. This is 
        handled automatically by the Client and enables faster response processing.
        Default: `False`

        :logprobs (optional): Include the log probabilities on the `logprobs` 
        most likely tokens, as well the chosen tokens. For example, if `logprobs` is 5, 
        the API will return a list of the 5 most likely tokens. The API will always 
        return the logprob of the sampled token, so there may be up to `logprobs+1` 
        elements in the response. The maximum value for `logprobs` is 5.
        Default: `None`

        :stop (optional): Up to 4 sequences where the API will stop generating 
        further tokens. The returned text will not contain the stop sequence.
        Default: `None`

        :presence_penalty (optional): Number between `-2.0` and `2.0`. Positive values 
        penalize new tokens based on whether they appear in the text so far, increasing the 
        model's likelihood to talk about new topics
        Default: `0.0`

        :frequency_penalty (optional): Number between `-2.0` and `2.0`. Positive 
        values penalize new tokens based on their existing frequency in the text so 
        far, decreasing the model's likelihood to repeat the same line verbatim.
        Default: `0.0`

        :logit_bias (optional): Modify the likelihood of specified tokens appearing in the completion.
        Accepts a json object that maps tokens (specified by their token ID in the GPT tokenizer) to an associated 
        bias value from -100 to 100. You can use this tokenizer tool (which works for both GPT-2 and GPT-3) to 
        convert text to token IDs. Mathematically, the bias is added to the logits generated by the model prior 
        to sampling. The exact effect will vary per model, but values between -1 and 1 should decrease or increase 
        likelihood of selection; values like -100 or 100 should result in a ban or exclusive selection of the 
        relevant token.
        As an example, you can pass `{"50256": -100}` to prevent the `<|endoftext|>` token from being generated.
        Default: `None`

        :user (optional): A unique identifier representing your end-user, which can help OpenAI to 
        monitor and detect abuse.

        :functions (optional): A list of dictionaries representing the functions to call
        
        :function_call (optional): The name of the function to call. Default: `auto` if functions are provided

        :response_format (optional): The format of the response. Default: `text`

        :seed (optional): An integer seed for random sampling. Must be between 0 and 2**32 - 1

        :tools (optional): A list of dictionaries representing the tools to use

        :tool_choice (optional): The name of the tool to use. Default: `auto` if tools are provided

        :auto_retry (optional): Whether to automatically retry the request if it fails due to a rate limit error.

        :auto_retry_limit (optional): The maximum number of times to retry the request if it fails due to a rate limit error.

        Default: `None`

        Returns: `ChatResponse`
        """
        return await super().async_create(
            input_object = input_object,
            parse_stream = parse_stream,
            timeout = timeout,
            auto_retry = auto_retry,
            auto_retry_limit = auto_retry_limit,
            header_cache_keys = header_cache_keys,
            **kwargs
        )

    



from __future__ import annotations

"""
The Base HTTP Client for KeyCloak
"""
import time
import httpx
import asyncio
import typing as t
import threading
from fastapi import HTTPException
from lzl.logging import logger
from .registry import kv_pdict, kv_available

if t.TYPE_CHECKING:
    from .manager import KeycloakManager


class KeycloakAuth(httpx.Auth):
    def __init__(self, client: 'KeycloakManager'):
        self.c = client
        self.token = None
        self.token_expiry = 0
        self._sync_lock = threading.RLock()
        self._async_lock = asyncio.Lock()
        if kv_available:
            self._set_from_kv()
    
    def _set_from_kv(self):
        """
        Sets the Keycloak token and expiry from the KV store.
        """
        try:
            if kv_pdict.contains('tokendata'):
                token_data: t.Dict[str, t.Any] = kv_pdict.get('tokendata')
                self.token = token_data.get('access_token')
                self.token_expiry = token_data.get('expires_at', 0)
        except Exception as e:
            logger.error(f"Failed to get Keycloak token from KV store: {e}")

    async def arefresh_token(self):
        """        
        Refresh the Keycloak token.
        """
        data = {
            "grant_type": "password",
            "client_id": self.c.client_id,
            "username": self.c.username,
            "password": self.c.password,
        }
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        async with httpx.AsyncClient(timeout = 60.0) as client:
            r = await client.post(self.c.token_url, data=data, headers=headers)
        if r.status_code != 200:
            logger.error(f"Failed to get Keycloak token: {r.text}")
            raise HTTPException(status_code=500, detail=f"Failed to get Keycloak token: {r.text}")
        token_data = r.json()
        logger.info(f"Keycloak token data: {token_data}")
        token_data['expires_at'] = time.time() + token_data.get("expires_in", 60)
        self.token = token_data["access_token"]
        self.token_expiry = token_data['expires_at']
        if kv_available: 
            logger.info("Saving Keycloak token to KV store")
            try:
                await kv_pdict.aset('tokendata', token_data, ex = (token_data.get("expires_in", 60) - 15))
            except Exception as e:
                logger.error(f"Failed to set Keycloak token in KV store: {e}")

    async def aget_token(self):
        """
        Fetch the Keycloak token, refreshing it if necessary.
        """
        async with self._async_lock:
            if self.token is None or time.time() > self.token_expiry - 30:
                logger.info(f"Refreshing Keycloak token: {time.time()} > {self.token_expiry - 30}")
                await self.arefresh_token()
        return self.token

    async def async_auth_flow(self, request: httpx.Request) -> t.AsyncIterator[httpx.Request]:
        token = await self.aget_token()
        request.headers["Authorization"] = f"Bearer {token}"
        yield request

    def refresh_token(self):
        """        
        Refresh the Keycloak token.
        """
        data = {
            "grant_type": "password",
            "client_id": self.c.client_id,
            "username": self.c.username,
            "password": self.c.password,
        }
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        with httpx.Client(timeout = 60.0) as client:
            r = client.post(self.c.token_url, data=data, headers=headers)
        if r.status_code != 200:
            logger.error(f"Failed to get Keycloak token: {r.text}")
            raise HTTPException(status_code=500, detail=f"Failed to get Keycloak token: {r.text}")
        token_data = r.json()
        self.token = token_data["access_token"]
        self.token_expiry = time.time() + token_data.get("expires_in", 60)
        if kv_available:
            try:
                kv_pdict.set('tokendata', token_data, ex = (token_data.get("expires_in", 60) - 30))
            except Exception as e:
                logger.error(f"Failed to set Keycloak token in KV store: {e}")

    def get_token(self):
        """
        Get the Keycloak token, refreshing it if necessary.
        """
        if self.token is None or time.time() > self.token_expiry - 30:
            self.refresh_token()
        return self.token
    
    def sync_auth_flow(self, request: httpx.Request) -> t.Iterator[httpx.Request]:
        token = self.get_token()
        request.headers["Authorization"] = f"Bearer {token}"
        yield request



from __future__ import annotations

"""
Pydantic Settings with additional helpers
"""
import os
import inspect
# import pkg_resources
from importlib import metadata
from pathlib import Path
from .base import (
    Field,
    BaseSettings as _BaseSettings,
    PYDANTIC_VERSION,
    get_pydantic_dict,
)
from .properties import eproperty
from typing import Optional, Dict, Any, List, TYPE_CHECKING

if PYDANTIC_VERSION == 2:
    from pydantic import PrivateAttr

if TYPE_CHECKING:
    from lzo.types import AppEnv

class BaseSettings(_BaseSettings):
    """
    Base Settings
    """

    if PYDANTIC_VERSION == 2:
        _extra: Dict[str, Any] = PrivateAttr(default_factory = dict)
    else:
        _extra: Dict[str, Any] = Field(default_factory = dict, exclude = True, hidden = True)

    
    class Config:
        env_prefix: str = ""
        case_sensitive: bool = False
        arbitrary_types_allowed: bool = True


    @property
    def __module_path(self) -> Path:
        """
        [V1 - Broken] Gets the module root path

        https://stackoverflow.com/questions/25389095/python-get-path-of-root-project-structure
        """
        p = Path(pkg_resources.get_distribution(self.module_name).location)
        if 'src' in p.name and p.joinpath(self.module_name).exists():
            p = p.joinpath(self.module_name)
        elif p.joinpath('src').exists() and p.joinpath('src', self.module_name).exists():
            p = p.joinpath('src', self.module_name)
        return p
    
    @eproperty
    def module_path(self) -> Path:
        """
        Gets the project root directory for a given module name using importlib.metadata.
        Corrected to properly find the package directory within site-packages.
        Replicates the logic of the original pkg_resources code, attempting to find
        a "project root" within the installation location, considering 'src' directories.

        Returns:
            Path: The project root path if found, otherwise None.
        """
        # sourcery skip: remove-redundant-path-exists
        try:
            dist = metadata.distribution(self.module_name)
            site_packages_path = Path(dist.locate_file('.'))  # Get site-packages path
            module_base_name = self.module_name.split('.')[0] # Handle submodules like pandas.plotting
            package_dir_path = site_packages_path / module_base_name # Construct package directory path
            if package_dir_path.exists() and package_dir_path.is_dir(): # Verify package dir exists
                p = package_dir_path
            else:
                # Fallback: if package dir not found (unusual but handle it), use site-packages
                p = site_packages_path

            if p.joinpath('src').exists() and p.joinpath('src', module_base_name).exists():
                p = p.joinpath('src', module_base_name)
            return p

        except metadata.PackageNotFoundError:
            return None

    @eproperty
    def module_config_path(self) -> Path:
        """
        Returns the config module path
        """
        return Path(inspect.getfile(self.__class__)).parent
    
    @eproperty
    def module_name(self) -> str:
        """
        Returns the module name
        """
        return self.__class__.__module__.split(".")[0]
    
    @eproperty
    def module_version(self) -> str:
        """
        Returns the module version
        """
        return metadata.version(self.module_name)
        # return pkg_resources.get_distribution(self.module_name).version
    
    @eproperty
    def module_pkg_name(self) -> str:
        """
        Returns the module pkg name
        
        {pkg}/src   -> src
        {pkg}/{pkg} -> {pkg}
        """
        config_path = self.module_config_path.as_posix()
        module_path = self.module_path.as_posix()
        return config_path.replace(module_path, "").strip().split("/", 2)[1]

    @eproperty
    def in_k8s(self) -> bool:
        """
        Returns whether the app is running in kubernetes
        """
        from lzo.utils.system import is_in_kubernetes
        return is_in_kubernetes()
    
    @eproperty
    def host_name(self) -> str:
        """
        Returns the hostname
        """
        from lzo.utils.system import get_host_name
        return get_host_name()

    def update_config(self, **kwargs):
        """
        Update the config for the other settings
        """
        for k, v in kwargs.items():
            if not hasattr(self, k): continue
            if isinstance(getattr(self, k), Path):
                setattr(self, k, Path(v))
            elif isinstance(getattr(self, k), self.__class__):
                val = getattr(self, k)
                if hasattr(val, 'update_config'):
                    val.update_config(**v)
                else: val = val.__class__(**v)
                setattr(self, k, val)
            else: 
                setattr(self, k, v)


    def set_envvars(self):
        """
        Update the Env variables for the current session
        """
        data = get_pydantic_dict(self, exclude_none=True)
        for k, v in data.items():
            if isinstance(v, BaseSettings):
                v.set_envvars()
            else:
                os.environ[self.Config.env_prefix + k.upper()] = str(v)
    

    def get(self, name, default: Any = None):
        """
        Get an attribute from the model
        """
        return getattr(self, name, default)


    @classmethod
    def _fetch_app_env(cls, module: Optional[str] = None) -> Optional['AppEnv']:
        """
        Fetches the App Env
        """
        if module is None: module = cls.__module__.split('.')[0]
        from lzo.types import AppEnv
        return AppEnv.from_module_name(module)



from __future__ import annotations

"""
Remote Object Storage Backend Dict-Like Persistence
"""

import re
import abc
import time
import datetime
import contextlib
from lzo.types import BaseModel, Field, model_validator, Literal, eproperty
from lzl.io.file import File
from typing import TypeVar, Generic, Any, Dict, Optional, Union, Tuple, Iterable, List, Type, Callable, Generator, AsyncGenerator, TYPE_CHECKING
from ..base import logger

if TYPE_CHECKING:
    from kvdb import Lock, AsyncLock
    from lzl.io.file import FileLike
    from .main import ObjStorageStatefulBackend


class ExpObject(BaseModel):
    """
    The Expiration Object
    """
    key: str
    ex: int
    expires_at: datetime.datetime
    
    @property
    def is_expired(self) -> bool:
        """
        Returns True if the Object is Expired

        - We only cache it if it is expired
        """
        if 'is_expired' not in self._extra and \
            self.expires_at < datetime.datetime.now(tz=datetime.timezone.utc):
                self._extra['is_expired'] = True
        return self._extra.get('is_expired', False)
    
    def set_new_exp(self, ex: int):
        """
        Sets a New Expiration
        """
        self.expires_at = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(seconds = ex)

    @classmethod
    def create(cls, key: str, ex: int):
        """
        Creates a new expiration object
        """
        return cls(key = key, ex = ex, expires_at = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(seconds = ex))


class ExpirationFile(BaseModel):
    """
    The Expiration File
    """
    index: Dict[str, ExpObject] = Field(default_factory = dict)

    def add_exp(self, key: str, ex: int):
        """
        Adds an Expiration
        """
        self.index[key] = ExpObject(key = key, ex = ex, expires_at = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(seconds = ex))


    def set_exp(self, key: str, ex: int):
        """
        Sets an Expiration
        """
        if key not in self.index:
            self.add_exp(key, ex)
        else:
            self.index[key].set_new_exp(ex)

    def set_exps(self, *keys: str, ex: int):
        """
        Sets an Expiration
        """
        for key in keys:
            self.set_exp(key, ex)

    def remove_exps(self, *keys: str):
        """
        Removes an Expiration
        """
        for key in keys:
            _ = self.index.pop(key, None)

    def get_expired_keys(self) -> List[str]:
        """
        Returns the Expired Keys
        """
        return [
            key
            for key, exp_obj in self.index.items()
            if exp_obj.is_expired
        ]

class ExpirationBackend(abc.ABC):
    """
    Expiration Backend
    """
    name: Optional[str] = None

    def __init__(self, backend: 'ObjStorageStatefulBackend'):
        """
        Initializes the Expiration Backend
        """
        self._extra: Dict[str, Any] = {}
        self.backend = backend
        self._setup_()
        self._handle_migration_()

    def _handle_migration_(self):
        """
        Handles the migration
        """
        pass
    
    def _setup_(self):
        """
        Sets up the expiration backend
        """
        pass
    
    @classmethod
    def is_available(cls) -> bool:
        """
        Returns whether the backend is available
        """
        return True
    

    def _check(self, *keys: str, validate: Optional[bool] = False, **kwargs):
        """
        Runs the expiration check

        Previously: `_run_expiration_check`
        """
        pass

    async def _acheck(self, *keys: str, validate: Optional[bool] = False, **kwargs):
        """
        Runs the expiration check

        Previously: `_arun_expiration_check`
        """
        pass


    def _set(self, *keys: str, ex: Optional[int] = None, validate: Optional[bool] = False, **kwargs):
        """
        Sets the expiration

        Previously: `_set_expiration`
        """
        pass

    async def _aset(self, *keys: str, ex: Optional[int] = None, validate: Optional[bool] = False, **kwargs):
        """
        Sets the expiration

        Previously: `_set_expiration`
        """
        pass

    def _remove(self, *keys: str, **kwargs):
        """
        Removes the expiration
        
        Previously: `_remove_expiration`
        """
        pass

    async def _aremove(self, *keys: str, **kwargs):
        """
        Removes the expiration
        
        Previously: `_remove_expiration`
        """
        pass


class FileExpirationBackend(ExpirationBackend):
    """
    File Expiration Backend
    """
    name: Optional[str] = 'file'

    def _setup_(self):
        """
        Sets up the expiration backend
        """
        from lzl.io.ser import get_serializer
        self.exp_file: 'File' = self.backend.base_key.joinpath(f'.{self.backend.name}.metadata.expires')
        self.ser = get_serializer(serializer = 'json')

    def _load(self, *args, **kwargs) -> 'ExpirationFile':
        """
        Loads the Expirations
        """
        if not self.exp_file.exists(): return ExpirationFile()
        try:
            return self.ser.loads(self.exp_file.read_text())
        except Exception as e:
            logger.error(f'Error Loading Expirations: {e}')
            return ExpirationFile()

    def _save(self, exps: 'ExpirationFile', **kwargs):
        """
        Saves the Expirations
        """
        self.exp_file.write_text(self.ser.dumps(exps))
    
    async def _aload(self, *args, **kwargs) -> 'ExpirationFile':
        """
        Loads the Expirations
        """
        if not await self.exp_file.aexists(): return ExpirationFile()
        try:
            return await self.ser.aloads(await self.exp_file.aread_text())
        except Exception as e:
            logger.error(f'Error Loading Expirations: {e}')
            return ExpirationFile()

    async def _asave(self, exps: 'ExpirationFile', **kwargs):
        """
        Saves the Expirations
        """
        await self.exp_file.awrite_text(await self.ser.adumps(exps))

    @contextlib.contextmanager
    def handler(self, *args, **kwargs) -> Generator['ExpirationFile', None, None]:
        """
        Handles the expiration file
        """
        exps = self._load()
        try:
            yield exps
        except Exception as e:
            logger.error(f'Error Handling Expirations: {e}')
        finally:
            self._save(exps)
            # logger.info(f'Saved Expirations: {exps}')
            
    @contextlib.asynccontextmanager
    async def ahandler(self, *args, **kwargs) -> AsyncGenerator['ExpirationFile', None]:
        """
        Handles the expiration file
        """
        exps = await self._aload()
        try:
            yield exps
        except Exception as e:
            logger.error(f'Error Handling Expirations: {e}')
        finally:
            await self._asave(exps)
            # logger.info(f'Saved Expirations: {exps}')

    def _check(self, *keys: str, **kwargs):
        """
        Runs the expiration check
        """
        # Maps to _arun_expiration_check
        if not self.exp_file.exists(): return
        with self.handler() as exps:
            expired_keys = exps.get_expired_keys()
            if expired_keys: 
                self.backend._clear(*expired_keys)
                exps.remove_exps(*keys)

    async def _acheck(self, *keys: str, **kwargs):
        """
        Runs the expiration check
        """
        # Maps to _arun_expiration_check
        # if self.exp_file is None: return
        if not await self.exp_file.aexists(): return
        async with self.ahandler() as exps:
            expired_keys = exps.get_expired_keys()
            if expired_keys: 
                await self.backend._aclear(*expired_keys)
                exps.remove_exps(*keys)

    def _set(self, *keys: str, ex: Optional[int] = None, validate: Optional[bool] = False, **kwargs):
        """
        Sets the expiration for the keys
        """
        # logger.info(f'Setting Expirations: {keys} -> {ex}')
        ex = ex or self.backend.expiration
        if ex is None: return
        with self.handler() as exps:
            exps.set_exps(*keys, ex = ex)
            if validate: 
                expired_keys = exps.get_expired_keys()
                if expired_keys: 
                    self.backend._clear(*expired_keys)
                    exps.remove_exps(*expired_keys)

    async def _aset(self, *keys: str, ex: Optional[int] = None, validate: Optional[bool] = False, **kwargs):
        """
        Sets the expiration
        """
        # logger.info(f'Setting Expirations: {keys} -> {ex}')
        ex = ex or self.backend.expiration
        if ex is None: return
        async with self.ahandler() as exps:
            exps.set_exps(*keys, ex = ex)
            if validate: 
                expired_keys = exps.get_expired_keys()
                if expired_keys: 
                    await self.backend._aclear(*expired_keys)
                    exps.remove_exps(*expired_keys)

    def _remove(self, *keys: str, **kwargs):
        """
        Removes the expiration
        """
        if not self.exp_file.exists(): return
        with self.handler() as exps:
            exps.remove_exps(*keys)

    async def _aremove(self, *keys: str, **kwargs):
        """
        Removes the expiration
        """
        if not await self.exp_file.aexists(): return
        async with self.ahandler() as exps:
            exps.remove_exps(*keys)


class RedisExpirationBackend(ExpirationBackend):
    """
    Redis Expiration Backend
    """
    name: Optional[str] = 'redis'
    _is_avail: Optional[bool] = None

    def _setup_(self):
        """
        Sets up the expiration backend
        """
        from kvdb import KVDBClient
        from lzl.io.ser import get_serializer
        from lzl.ext.aiorun import create_task
        from lzo.utils.aioexit import register
        
        # >>> s3://mybucket/mykey/file.metadata.expires
        # >>> mybucket.mykey.file.metadata:expirations
        self.create_task = create_task
        self.keybase = self.backend.base_key.as_posix().split('://', 1)[-1].replace('/', '.')
        self.exp_base_key = f'_fexp_:{self.keybase}'
        self.ser = get_serializer(serializer = 'json')
        self.session = KVDBClient.get_session(
            name = f'{self.backend.name}.expirations',
            serializer = None,
            decode_responses = True, 
            # serializer = 'json',
        )
        # register(self._finalize_)
    
    @eproperty
    def lock(self) -> 'Lock':
        """
        Returns the lock
        """
        return self.session.lock(
            f'{self.keybase}:lock',
            timeout = 60 * 5,
            blocking = True,
        )
    
    @eproperty
    def alock(self) -> 'AsyncLock':
        """
        Returns the async lock
        """
        return self.session.alock(
            f'{self.keybase}:lock',
            timeout = 60 * 5,
            blocking = True,
        )

        
    def _handle_migration_(self):
        """
        Handles the migration
        """
        # Check whether the key exists
        with self.lock:
            if self.session.exists(self.exp_base_key): return

            exp_file: 'File' = self.backend.base_key.joinpath(f'.{self.backend.name}.metadata.expires')
            if not exp_file.exists(): return

            logger.info(f'Migrating Expirations from {exp_file} to {self.exp_base_key} (File -> Redis)')
            exps: 'ExpirationFile' = self.ser.loads(exp_file.read_text())
            set_data = {key: exp.expires_at.timestamp() for key, exp in exps.index.items()}

            self.session.hmset(self.exp_base_key, set_data)
            logger.info(f'Migrated `{len(set_data)}` Expirations from {exp_file} to {self.exp_base_key} (File -> Redis)')
        
        # exp_file.unlink()

    @classmethod
    def is_available(cls) -> bool:
        """
        Returns whether the backend is available
        """
        if cls._is_avail is not None: return cls._is_avail
        try:
            import os
            import kvdb
            import socket
            from lzo.utils.helpers.envvars import is_in_ci_env
            if not os.getenv('REDIS_URL') or is_in_ci_env(): return False
            # Check that the host is available
            redis_url = os.getenv('REDIS_URL').split('://', 1)[-1]
            if '@' in redis_url:
                redis_url = redis_url.split('@', 1)[-1]
            if ':' in redis_url:
                redis_url = redis_url.split(':', 1)[0]
            socket.gethostbyname(redis_url)
            # socket.gethostbyname(os.getenv('REDIS_URL').split('://', 1)[-1].split(':', 1)[0])
            cls._is_avail = True
            return True
        
        except ImportError as e:
            cls._is_avail = False
            return False
        
        except Exception as e:
            logger.error(f'Error validating Redis Availability: {e}')
            return False


    def _validate(self, *args, ignore_keys: Iterable[str] = None, **kwargs):
        """
        Validates the expired keys
        """
        expired_keys = []
        with self.lock:
            now = time.time()
            if not self.session.exists(self.exp_base_key): return
            exps = self.session.hgetall(self.exp_base_key)
            for key, exp in exps.items():
                if ignore_keys and key in ignore_keys: continue
                try:
                    exp = float(exp)
                    if exp <= now:
                        expired_keys.append(key)
                        continue
                except Exception as e:
                    logger.error(f'Error Parsing Expiration: {e} - {key}')
                    expired_keys.append(key)
                    continue
            
        if expired_keys: 
            self.session.hdel(self.exp_base_key, *expired_keys)
            self.backend._clear(*expired_keys)

    async def _avalidate(self, *args, ignore_keys: Iterable[str] = None, **kwargs):
        """
        Validates the expired keys
        """
        expired_keys = []
        async with self.alock:
            now = time.time()
            if not await self.session.aexists(self.exp_base_key): return
            exps = await self.session.ahgetall(self.exp_base_key)
            for key, exp in exps.items():
                if ignore_keys and key in ignore_keys: continue
                try:
                    exp = float(exp)
                    if exp <= now:
                        expired_keys.append(key)
                        continue
                except Exception as e:
                    logger.error(f'Error Parsing Expiration: {e} - {key}')
                    expired_keys.append(key)
                    continue
        if expired_keys: 
            await self.session.ahdel(self.exp_base_key, *expired_keys)
            await self.backend._aclear(*expired_keys)

    def _check(self, *keys: str, validate: Optional[bool] = False, **kwargs):
        """
        Runs the expiration check

        Previously: `_run_expiration_check`
        """
        # with self.lock:
        if not self.session.exists(self.exp_base_key): return
        expired_keys = []
        now = time.time()
        for key in keys:
            if self.session.hexists(self.exp_base_key, key):
                try:
                    exp = float(self.session.hget(self.exp_base_key, key))
                    if exp <= now:
                        expired_keys.append(key)
                        continue
                except Exception as e:
                    logger.error(f'Error Parsing Expiration: {e} - {key}')
                    expired_keys.append(key)
                    continue
        
        # if keys: self.session.hdel(self.exp_base_key, *keys)
        if expired_keys: 
            self.session.hdel(self.exp_base_key, *expired_keys)
            self.backend._clear(*expired_keys)
        if validate: self._validate(ignore_keys=keys)
    

    async def _acheck(self, *keys: str, validate: Optional[bool] = False, **kwargs):
        """
        Runs the expiration check

        Previously: `_arun_expiration_check`
        """
        # async with self.alock:
        if not await self.session.aexists(self.exp_base_key): return
        expired_keys = []
        now = time.time()
        for key in keys:
            if await self.session.ahexists(self.exp_base_key, key):
                try:
                    exp = float(await self.session.ahget(self.exp_base_key, key))
                    if exp <= now:
                        expired_keys.append(key)
                        continue
                except Exception as e:
                    logger.error(f'Error Parsing Expiration: {e} - {key}')
                    expired_keys.append(key)
                    continue
        if expired_keys: 
            await self.session.ahdel(self.exp_base_key, *expired_keys)
            await self.backend._aclear(*expired_keys)

            # if keys: await self.session.ahdel(self.exp_base_key, *keys)
        if validate: await self._avalidate(ignore_keys = keys)

    def _set(self, *keys: str, ex: Optional[int] = None, validate: Optional[bool] = False, **kwargs):
        """
        Sets the expiration

        Previously: `_set_expiration`
        """
        ex = ex or self.backend.expiration
        if ex is None: return
        with self.lock:
            exp = time.time() + ex
            mapping = {
                key: exp for key in keys
            } 
            self.session.hmset(self.exp_base_key, mapping)
        if validate: self._validate(ignore_keys = keys)

    async def _aset(self, *keys: str, ex: Optional[int] = None, validate: Optional[bool] = False, **kwargs):
        """
        Sets the expiration

        Previously: `_set_expiration`
        """
        ex = ex or self.backend.expiration
        if ex is None: return
        async with self.alock:
            exp = time.time() + ex
            mapping = {
                key: exp for key in keys
            } 
            await self.session.ahmset(self.exp_base_key, mapping)
        if validate: await self._avalidate(ignore_keys = keys)

    def _remove(self, *keys: str, **kwargs):
        """
        Removes the expiration
        
        Previously: `_remove_expiration`
        """
        if not self.session.exists(self.exp_base_key): return
        self.session.hdel(self.exp_base_key, *keys)

    async def _aremove(self, *keys: str, **kwargs):
        """
        Removes the expiration
        
        Previously: `_remove_expiration`
        """
        if not await self.session.aexists(self.exp_base_key): return
        await self.session.ahdel(self.exp_base_key, *keys)
            



from __future__ import annotations

"""
Temporal Registry Methods
"""
import copy
import functools
import dataclasses
import typing as t
from temporalio import workflow
from temporalio import activity
from ..utils import logger

if t.TYPE_CHECKING:
    from ..mixins import (
        BaseTemporalMixin, TemporalWorkflowMixin, TemporalActivityMixin,
        TemporalMixinT, TemporalWorkflowT, TemporalActivityT, TemporalDispatchT,
        MixinKinds
    )


@dataclasses.dataclass
class WorkflowConfigState:
    set_run: t.Optional[bool] = False
    set_signal: t.Optional[bool] = False
    set_query: t.Optional[bool] = False

    skip_set_signal: t.Optional[bool] = False
    skip_set_query: t.Optional[bool] = False


def configure_workflow_from_config(
    obj: t.Type['TemporalWorkflowT'],
    state: WorkflowConfigState,
    **kwargs,
):
    # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
    """
    Configures the Workflow from config
    """
    if obj.config.get('run'):
        state.set_run = True
        run_func = getattr(obj, obj.config['run'])
        run_func = workflow.run(run_func)
        setattr(obj, obj.config['run'], run_func)
    
    if obj.config.get('signal'):
        state.set_signal = True
        sig_config = obj.config['signal']
        if isinstance(sig_config, str):
            sig_func_name = sig_config
            sig_func = getattr(obj, sig_config)
            sig_config = {}
        elif isinstance(sig_config, dict):
            # Single Function
            if sig_config.get('func'):
                sig_func_name = sig_config.pop('func')
                sig_func = getattr(obj, sig_func_name)
            
            # Multiple Functions
            else:
                state.skip_set_signal = True
                for sig_func_name, func_config in sig_config.items():
                    sig_func = getattr(obj, sig_func_name)
                    if func_config.get('unfinished_policy') and isinstance(func_config['unfinished_policy'], str):
                        func_config['unfinished_policy'] = workflow.HandlerUnfinishedPolicy[func_config['unfinished_policy']]
                    sig_func = workflow.signal(sig_func, **func_config)
                    setattr(obj, sig_func_name, sig_func)

        # Support Tuple and Lists later
        else:
            logger.warning(f'Unsupported Signal Config Type: {sig_config} ({type(sig_config)})', prefix = 'Temporal', colored = True)
            state.skip_set_signal = True
        
        if not state.skip_set_signal:
            if sig_config.get('unfinished_policy') and isinstance(sig_config['unfinished_policy'], str):
                sig_config['unfinished_policy'] = workflow.HandlerUnfinishedPolicy[sig_config['unfinished_policy']]
            sig_func = workflow.signal(sig_func, **sig_config)
            setattr(obj, sig_func_name, sig_func)

    if obj.config.get('query'):
        state.set_query = True
        query_config = obj.config['query']
        if isinstance(query_config, str):
            query_func_name = query_config
            query_func = getattr(obj, query_config)
            query_config = {}
        
        elif isinstance(query_config, dict):
            # Single Function
            if query_config.get('func'):
                query_func_name = query_config.pop('func')
                query_func = getattr(obj, query_func_name)
            
            # Multiple Functions
            else:
                state.skip_set_query = True
                for query_func_name, func_config in query_config.items():
                    query_func = getattr(obj, query_func_name)
                    query_func = workflow.query(query_func, **func_config)
                    setattr(obj, query_func_name, query_func)
        
        # Support Tuple and Lists later
        else:
            logger.warning(f'Unsupported Query Config Type: {query_config} ({type(query_config)})', prefix = 'Temporal', colored = True)
            state.skip_set_query = True
        
        if not state.skip_set_query:
            query_func = workflow.query(query_func, **query_config)
            setattr(obj, query_func_name, query_func)

def create_run_wrapper(
    obj: t.Type['TemporalWorkflowT'],
    run_func: t.Callable[..., t.Any],
) -> t.Callable[..., t.Any]:
    """
    Creates a run wrapper
    """
    @functools.wraps(run_func)
    def wrapper(*args, **kwargs):
        """
        Runs the run function
        """
        return run_func(*args, **kwargs)
    
    # Clone the method into the child object
    if obj.__qualname__ not in run_func.__qualname__:
        wrapper.__qualname__ = f'{obj.__qualname__}.run'
        setattr(obj, 'run', wrapper)
    return wrapper

def configure_workflow(
    obj: t.Type['TemporalWorkflowT'],
    **kwargs,
):
    # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
    """
    Configures the Workflow
    """
    if obj.enable_init:
        init_func = getattr(obj, '__init__')
        init_func = workflow.init(init_func)
        setattr(obj, '__init__', init_func)
    
    state = WorkflowConfigState()
    obj._configure_cls_()
    obj._configure_display_name_()
    obj._configure_gen_id_()
    if obj.config: configure_workflow_from_config(obj, state)
    if not state.set_run and hasattr(obj, 'run'):
        # if obj.__qualname__ not in getattr(obj, 'run').__qualname__:
        #     # import types
        #     # Clone the method into the child object
        #     # obj.run = types.MethodType(obj.run, obj) 
        #     obj.run = create_run_wrapper(obj, obj.run)

            # f = getattr(obj, 'run')
            # f.__qualname__ = f'{obj.__qualname__}.run'
            # setattr(obj, 'run', f)

        run_func = getattr(obj, 'run')
        # Clone the method into the child object
        # if obj.__qualname__ not in run_func.__qualname__:
        # logger.info(f'Running Workflow: {run_func} - {run_func.__qualname__} - {obj.__qualname__}')
        run_func = workflow.run(run_func)
        setattr(obj, 'run', run_func)

    if not state.set_signal and hasattr(obj, 'signal'):
        sig_func = getattr(obj, 'signal')
        sig_func = workflow.signal(sig_func)
        setattr(obj, 'signal', sig_func)
    
    if not state.set_query and hasattr(obj, 'query'):
        query_func = getattr(obj, 'query')
        query_func = workflow.query(query_func)
        setattr(obj, 'query', query_func)
    
    obj._on_register_hook_()



@dataclasses.dataclass
class ActivityConfigState:
    set_funcs: t.Optional[bool] = False


def configure_activity_from_config(
    obj: t.Type['TemporalActivityT'],
    state: ActivityConfigState,
    **kwargs,
):
    # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
    """
    Configures the Activity from config
    """
    from .main import activity_defn
    if obj.config.get('func'):
        state.set_funcs = True
        func = obj.config['func']
        if isinstance(func, str):
            func_name = func
            act_func = getattr(obj, func)
            act_conf = {}
        else:
            # It's a dict
            func_name = list(func.keys())[0]
            act_func = getattr(obj, func_name)
            act_conf = dict(func.values())

        act_func = activity_defn(obj, act_func, **act_conf)
        setattr(obj, func_name, act_func)
    
    elif obj.config.get('funcs'):
        state.set_funcs = True
        funcs = obj.config['funcs']
        for func_name, func_conf in funcs.items():
            act_func = getattr(obj, func_name)
            act_func = activity_defn(obj, act_func, **func_conf)
            setattr(obj, func_name, act_func)
    
def configure_activity(
    obj: t.Type['TemporalActivityT'],
    **kwargs,
):
    # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
    """
    Configures the Activity
    """
    state = ActivityConfigState()
    obj._configure_cls_()
    obj._configure_display_name_()
    obj._configure_gen_id_()
    if obj.config: configure_activity_from_config(obj, state)

    if not state.set_funcs:
        from .main import activity_defn
        # Detect all functions
        for name in dir(obj):
            # logger.info(name, prefix = obj.display_name, colored = True)
            if name.startswith('_') or name.endswith('_') or name in {'configure_registered', 'get_next_cron_run', 'generate_id'}: continue
            if obj.include_funcs and name not in obj.include_funcs: continue
            if obj.exclude_funcs and name in obj.exclude_funcs: continue
            func = getattr(obj, name)
            if not callable(func): continue
            logger.info(f'Registered: `|g|{obj.display_name}|e|:{name}`', prefix = 'Temporal', colored = True)
            act_func = activity_defn(obj, func)
            setattr(obj, name, act_func)
    
    obj._on_register_hook_()

def configure_dispatch(
    obj: t.Type['TemporalDispatchT'],
    **kwargs,
):
    """
    Configure the Dispatch
    """
    obj._configure_cls_()
    obj._on_preregister_hook_()
    from lzl.load import lazy_import
    if obj.input_type and isinstance(obj.input_type, str):
        obj.input_type = lazy_import(obj.input_type)
    if obj.result_type and isinstance(obj.result_type, str):
        obj.result_type = lazy_import(obj.result_type)
    obj._on_register_hook_()






from __future__ import annotations

"""
TemporalIO Config
"""
import os
import typing as t
from urllib.parse import urlparse
from lzo.types import BaseSettings, model_validator, Field
from lzl.types import eproperty
from lzl.proxied import ProxyObject
from lzl.io import File
from lzl.load import lazy_import
from lzo.utils.helpers.envvars import replace_envvars_in_text
from ..utils import logger
from .worker import TemporalWorkerConfig
from .telemetry import TemporalTelemetrySettings

if t.TYPE_CHECKING:
    from lzl.io.persistence import TemporaryData
    from temporalio.converter import DataConverter
    from temporalio.worker import Interceptor
    from temporalio.runtime import Runtime, TelemetryConfig
    from lzl.ext.temporal.loop.config import RuntimeConfig
    from lzl.ext.temporal.client import TemporalClient
    from ..registry import TemporalRegistry


class TemporalSettings(BaseSettings):
    """
    Temporal Settings
    """

    url: t.Optional[str] = Field(None, description = "The Temporal URL. Use a full URL including the protocol so that it can be parsed correctly, such as `https://localhost:7233`")
    host: t.Optional[str] = Field(None, description = "The Temporal host. Use a FQDN with the port such as `localhost:7233`")
    api_key: t.Optional[str] = Field(None, description = "The Temporal API Key")
    namespace: t.Optional[str] = Field('default', description = "The Temporal namespace")
    tls: t.Optional[bool] = Field(None, description = "If false, the default, do not use TLS. If true, use system default TLS configuration. If TLS configuration present, that TLS configuration will be used.")
    identity: t.Optional[str] = Field(None, description = "Identity for this client. If unset, a default is created based on the version of the SDK.")
    
    interceptors: t.List[str] = Field(default_factory=list, description = "The interceptors to use for the client")
    converter: t.Optional[str] = Field("serialized", description = "The converter to use for the client")

    default_factory: t.Optional[str] = Field(default="lzl.ext.temporal.loop.worker:WorkerFactory", description = "The worker factory to use for creating workers")
    pre_init: t.List[str] = Field(default_factory=list, description = "The pre-init functions to run before the worker is initialized")

    max_concurrent_activities: int = Field(default=100, description = "The maximum number of concurrent activities to run")
    max_concurrent_workflow_tasks: int = Field(default=100, description = "The maximum number of concurrent workflow tasks to run")
    disable_eager_activity_execution: bool = Field(default=True, description = "If true, the worker will not eagerly execute activities. This can improve performance for large workflows.")
    # metric_bind_address: str = Field(default="0.0.0.0:9000", description = "The address to bind the metrics server to")

    lazy: t.Optional[bool] = Field(None, description = "If true, the client will not connect until the first call is attempted. If false, the client will connect immediately.")
    default_task_queue: t.Optional[str] = Field(None, description = "The default task queue to use for this client.")

    data_jsonlib: t.Optional[str] = Field(None, description = "The JSON library to use for data conversion")
    data_compression: t.Optional[str] = Field(None, description = "The compression to use for data conversion")
    data_compression_level: t.Optional[int] = Field(None, description = "The compression level to use for data conversion")

    workers: t.List[TemporalWorkerConfig] = Field(default_factory=list, description = "The workers to use for the client")
    config_file: t.Optional[File] = Field(None, description = "The config file to load the settings from")
    client_class: t.Optional[str] = Field(None, description = "The client class to use")

    @model_validator(mode = 'after')
    def validate_settings(self):
        """
        Validates the settings
        """
        if self.config_file is not None:
            self._load_from_config_file_(self.config_file)
        if self.url is not None:
            self._construct_endpoint_()
        return self
    
    @eproperty
    def telemetry(self) -> TemporalTelemetrySettings:
        """
        Returns the Telemetry Settings
        """
        return TemporalTelemetrySettings()

    def _construct_endpoint_(self):
        """
        Constructs the URL Endpoint
        """
        url = urlparse(self.url)
        self.host = f'{url.hostname}:'
        if url.port: self.host += str(url.port)
        elif url.scheme == 'http': self.host += '80'
        elif url.scheme == 'https': self.host += '443'
        elif url.scheme in {'grpc', 'grpcs'}: self.host += '7233'
        else: raise ValueError(f'Invalid URL: {self.url} with Unknown Scheme: {url.scheme}')
        if url.scheme not in {'https', 'grpcs'}: self.tls = False
        if url.password or url.username: self.api_key = url.password or url.username
        if not self.tmpdata.has_logged(f'tmprl:settings:endpoint:{self.host}'):
            msg = ''
            if self.tls: msg += '|g|TLS|e|: Enabled'
            if self.api_key: msg += ', |g|Auth|e|: Enabled'
            if msg: msg = f" ({msg.lstrip(', ')})"    
            logger.info(f'Endpoint: `|g|{self.host}|e|`{msg}', prefix = 'Temporal', colored = True)
    
    def _handle_config_data_pre_(self, config_data: t.Dict[str, t.Any]):
        """
        Passthrough function to handle config data before it is loaded
        """
        pass
        
    def _handle_config_data_post_(self, config_data: t.Dict[str, t.Any]):
        """
        Passthrough function to handle config data after it is loaded
        """
        pass

    def _load_from_config_file_(self, config_file: File):
        """
        Loads the settings from a config file
        """
        import yaml
        config_file = File(config_file)
        config_text = replace_envvars_in_text(config_file.read_text())
        config_data: t.Dict[str, t.Any] = yaml.safe_load(config_text)
        self._handle_config_data_pre_(config_data)
        if config_data.get('workers'):
            for worker in config_data.pop('workers'):
                self.workers.append(TemporalWorkerConfig(**worker))
        elif config_data.get('worker'):
            self.workers.append(TemporalWorkerConfig(**config_data.pop('worker')))
        if config_data.get('telemetry'):
            self.telemetry.update(config_data.pop('telemetry'))
        self._handle_config_data_post_(config_data)
        for k,v in config_data.items():
            if hasattr(self, k): setattr(self, k, v)


    @property
    def client_kwargs(self) -> t.Dict[str, t.Any]:
        """
        Returns the client kwargs
        """
        kws = {
            'target_host': self.host,
            'namespace': self.namespace,
            'api_key': self.api_key,
            'tls': self.tls,
            'identity': self.identity,
            'lazy': self.lazy,
        }
        return {k: v for k,v in kws.items() if v is not None}
    
    def get_data_converter(
        self,
        converter: t.Optional[str] = None,
        jsonlib: t.Optional[str] = None,
        compression: t.Optional[str] = None,
        compression_level: t.Optional[int] = None,
        **kwargs,
    ) -> 'DataConverter':
        """
        Returns the data converter
        """
        from lzl.ext.temporal.io import get_default_data_converter
        return get_default_data_converter(
            converter = converter,
            jsonlib = jsonlib,
            compression = compression,
            compression_level = compression_level,
            **kwargs,
        )

    @eproperty
    def data_converter(self) -> 'DataConverter':
        """
        Returns the data converter
        """
        return self.get_data_converter(
            converter = self.converter,
            jsonlib = self.data_jsonlib,
            compression = self.data_compression,
            compression_level = self.data_compression_level,
            config = self,
        )
    
    @eproperty
    def tmpdata(self) -> 'TemporaryData':
        """
        Returns the temporary data
        """
        from lzl.io.persistence import TemporaryData
        return TemporaryData.from_module('lzl')
    
    @eproperty
    def registry(self) -> 'TemporalRegistry':
        """
        Returns the Temporal Registry
        """
        from lzl.ext.temporal.registry import registry
        return registry
    
    @eproperty
    def client_cls(self) -> t.Type['TemporalClient']:
        """
        Returns the Temporal Client Class
        """
        if self.client_class: return lazy_import(self.client_class)
        from lzl.ext.temporal.client import TemporalClient
        return TemporalClient
    
    def load_functions(self, *funcs: t.Sequence[t.Union[str, t.Type[t.Any]]], initialize: t.Optional[bool] = None) -> t.List[t.Any]:
        """
        Loads the functions
        """
        _funcs = []
        for func in funcs:
            if isinstance(func, str): func = lazy_import(func)
            if initialize: func = func()
            _funcs.append(func)
        return _funcs

    def build_runtime(
        self,
        runtime: t.Optional[str | t.Type['Runtime']] = None,
        telemetry: t.Optional['TelemetryConfig'] = None,
        **kwargs,
    ) -> t.Optional['Runtime']:
        """
        Builds the Runtime
        """
        if not telemetry and not runtime: return None
        if not runtime:
            from temporalio.runtime import Runtime
            runtime = Runtime
        elif isinstance(runtime, str): runtime = lazy_import(runtime)
        return runtime(telemetry = telemetry, **kwargs)

    async def get_client(
        self,
        host: t.Optional[str] = None,
        api_key: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        converter: t.Optional[str] = None,
        interceptors: t.Sequence[t.Union[t.Type['Interceptor'], str]] | None = None,
        runtime: t.Optional[t.Type['Runtime'] | str] = None,
        telemetry: t.Optional[bool | str] = None,
        telemetry_config: t.Optional[t.Dict[str, t.Any]] = None,
        telemetry_init_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs,
    ) -> 'TemporalClient':
        """
        Returns the Temporal Client
        """
        if not host: host = self.host
        if not api_key and self.api_key: api_key = self.api_key
        if not namespace: namespace = self.namespace
        if not converter: converter = self.data_converter
        else: converter = self.get_data_converter(converter = converter)
        
        if interceptors: interceptors = self.load_functions(*interceptors, initialize = True)
        else: interceptors = []

        if telemetry is None or (isinstance(telemetry, str) or telemetry is True):
            telemetry_config, telemetry_interceptor = self.telemetry.get_telemetry_config(
                provider = telemetry,
                init_kwargs = telemetry_init_kwargs,
                config_kwargs = telemetry_config,
                init_interceptor = True,
            )
            if telemetry_config: kwargs['runtime'] = self.build_runtime(runtime = runtime, telemetry = telemetry_config)
            if telemetry_interceptor: interceptors.append(telemetry_interceptor)

        return await self.client_cls.connect(
            host,
            api_key = api_key,
            namespace = namespace,
            data_converter = converter,
            interceptors = interceptors,
            **kwargs,
        )

    
    def get_runtime_config(
        self, 
        config_file: t.Optional[str | t.List[str]] = None,
        host: t.Optional[str] = None,
        name: t.Optional[str] = None,
        queue: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        identity: t.Optional[str] = None,
        activities: t.Optional[t.List[str]] = None,
        workflows: t.Optional[t.List[str]] = None,
        interceptors: t.Optional[t.List[str]] = None,
        converter: t.Optional[str] = None,
        telemetry: t.Optional[bool | str] = None,
        telemetry_config: t.Optional[t.Dict[str, t.Any]] = None,
        telemetry_init_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs,
    ) -> 'RuntimeConfig':
        """
        Returns the Runtime Config
        """
        if config_file is not None: 
            logger.info(f'Loading Runtime Config from Config File: {config_file}')
            if not isinstance(config_file, list): config_file = [config_file]
            for cf in config_file:
                self._load_from_config_file_(cf)
        if activities and workflows:
            _worker_config = dict(
                name = name or 'default-worker',
                queue = queue,
                identity = identity,
                namespace = namespace,
                activities = activities,
                workflows = workflows,
                **kwargs,
            )
            _worker_config = {k:v for k,v in _worker_config.items() if v is not None}
            new_worker = TemporalWorkerConfig(
                **_worker_config,
            )
            self.workers.append(new_worker)
            
        elif queue is not None: 
            for worker in self.workers:
                worker.queue = queue
        
        converter = self.get_data_converter(converter = converter) if converter else self.data_converter
        interceptors = interceptors or self.interceptors or []
        
        if telemetry is None or (isinstance(telemetry, str) or telemetry is True):
            telemetry_config, telemetry_interceptor = self.telemetry.get_telemetry_config(
                provider = telemetry,
                init_kwargs = telemetry_init_kwargs,
                config_kwargs = telemetry_config,
            )
            if telemetry_config: kwargs['telemetry'] = telemetry_config
            if telemetry_interceptor: interceptors.append(telemetry_interceptor)

        base_kwargs = dict(
            host = host or self.host,
            namespace = namespace or self.namespace,
            factory = self.default_factory,
            interceptors = interceptors,
            converter = converter,
            pre_init = self.pre_init,
            max_concurrent_activities = self.max_concurrent_activities,
            max_concurrent_workflow_tasks = self.max_concurrent_workflow_tasks,
            # metric_bind_address = self.metric_bind_address,
            workers = self.workers,
        )

        if kwargs: base_kwargs.update(kwargs)
        from lzl.ext.temporal.loop.config import RuntimeConfig
        return RuntimeConfig(**base_kwargs)

    class Config(BaseSettings.Config):
        env_prefix = "TEMPORAL_"



_tmprl_settings: t.Optional[TemporalSettings] = None
_tmprl_settings_cls: t.Optional[t.Type[TemporalSettings]] = TemporalSettings

def get_temporal_settings(**kwargs) -> TemporalSettings:
    """
    Returns the Temporal Settings
    """
    global _tmprl_settings
    if _tmprl_settings is None:
        _tmprl_settings = _tmprl_settings_cls(**kwargs)
    return _tmprl_settings

def set_temporal_settings_cls(cls: t.Type[TemporalSettings]):
    """
    Sets the Temporal Settings Class
    """
    global _tmprl_settings_cls
    _tmprl_settings_cls = cls
    
def set_temporal_settings(settings: TemporalSettings):
    """
    Sets the Temporal Settings
    """
    global _tmprl_settings
    # logger.info(f'Setting Temporal Settings: {settings}', prefix = 'Temporal', colored = True)
    _tmprl_settings = settings


config: TemporalSettings = ProxyObject(
    # obj_cls = TemporalSettings
    obj_getter = get_temporal_settings,
)


from __future__ import annotations


import typing as t
import functools
from temporalio.api.common.v1 import Payload
from temporalio.converter import (
    CompositePayloadConverter,
    DataConverter,
    DefaultPayloadConverter,
    JSONPlainPayloadConverter,
)
from .utils import get_json_meta, _json, is_serializable


class PydanticSerializedPayloadConverter(JSONPlainPayloadConverter):
    """Pydantic JSON payload converter.

    This extends the :py:class:`JSONPlainPayloadConverter` to override
    :py:meth:`to_payload` using the Pydantic encoder.
    """

    def to_payload(self, value: t.Any) -> t.Optional[Payload]:
        """Convert all values with Pydantic encoder or fail.

        Like the base class, we fail if we cannot convert. This payload
        converter is expected to be the last in the chain, so it can fail if
        unable to convert.
        """
        # We let JSON conversion errors be thrown to caller
        if not is_serializable(value):
            raise RuntimeError(f"Cannot serialize value of type {type(value)} - {value}")
        return Payload(
            metadata = get_json_meta(self.encoding, "serialized"),
            data = _json.dumps(
                value, 
                separators = (",", ":"), 
                sort_keys = True, 
            ),
        )
    
    def from_payload(self, payload: Payload, type_hint: t.Optional[t.Type] = None) -> t.Any:
        """
        Converts the payload to the type hint
        """
        return _json.loads(payload.data)


class PydanticPayloadConverter(CompositePayloadConverter):
    """Payload converter that replaces Temporal JSON conversion with Pydantic
    JSON conversion.
    """

    def __init__(self) -> None:
        super().__init__(
            *(
                PydanticSerializedPayloadConverter() if isinstance(c, JSONPlainPayloadConverter) else c
                for c in DefaultPayloadConverter.default_encoding_payload_converters
            )
        )



pydantic_data_converter = DataConverter(
    payload_converter_class = PydanticPayloadConverter
)
"""Data converter using Pydantic JSON conversion."""


from __future__ import annotations

"""
SQLite Backend Dict-Like Persistence
"""
import copy
import typing as t
from typing import TypeVar, Generic, Any, Dict, Optional, Union, Tuple, Iterable, List, Type, Callable, Generator, AsyncGenerator, TYPE_CHECKING
from ..base import BaseStatefulBackend, SchemaType, create_unique_id, logger
from .db import SqliteDB, SqliteDsn, Optimization, ENOVAL


class SqliteStatefulBackend(BaseStatefulBackend):
    """
    Implements an Sqlite Stateful Backend
    """

    name: Optional[str] = "sqlite"
    table: Optional[str] = "cache"
    expiration: Optional[int] = None
    auto_delete_invalid: Optional[bool] = False
    optimization: Optional[Optimization] = 'cache'

    def __init__(
        self,
        base_key: Optional[Union[str, SqliteDsn]] = None,
        name: Optional[str] = None,
        table: Optional[str] = None,
        expiration: Optional[int] = None,
        async_enabled: Optional[bool] = False,
        auto_delete_invalid: Optional[bool] = None,
        serializer: Optional[str] = 'json',
        serializer_kwargs: Optional[Dict[str, Any]] = None,
        optimization: Optional[str] = None,
        db_settings: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        Initializes the backend

        Args:
            base_key (str, SqliteDsn): 
                The base key for the backend.
                Either a string or a SQLite DSN.

                Examples:
                - base_key = 'sqlite://local/path/to/sqlite.db' - Will use the default table
                - base_key = 'sqlite://local/path/to/sqlite.db?table=mytable' - Will use the specified table
                - base_key = 'sqlite://host:1234?table=mytable' - Will use the specified table

            name (str, optional): The name of the backend. Defaults to None.
            table (str, optional): The table to use. Defaults to None.
            expiration (int, optional): The expiration time in seconds. Defaults to None.
        """
        # If it exists and is passed, then use it
        conn_uri: t.Optional[str] = kwargs.pop('conn_uri', None)
        is_remote: t.Optional[bool] = kwargs.pop('is_remote', None)
        _db: t.Optional[SqliteDB] = kwargs.pop('_db', None)

        configured = conn_uri is not None or _db is not None
        
        # We ignore the base key since each base key is a different table
        if not conn_uri:
            base_key_uri = SqliteDsn(base_key) if isinstance(base_key, str) else base_key
        
            # This is a local file
            if base_key_uri.path.endswith('.db'):
                conn_uri = str(base_key).split('?', 1)[0]
                conn_uri = conn_uri.replace('sqlite://', '')
                is_remote = False
            else:
                host = base_key_uri.hosts()[0]
                conn_uri = 'sqlite://'
                if host['username'] and host['password']:
                    conn_uri += f'{host["username"]}:{host["password"]}@'
                elif host['password']:
                    conn_uri += f'{host["password"]}@'
                conn_uri += f'{host["host"]}'
                if host['port']:
                    conn_uri += f':{host["port"]}'
                is_remote = True
            for k, v in base_key_uri.query_params():
                # If the arguments are passed, then prioritize them over the query params
                if k == 'table' and not table: table = v
                if k == 'expiration' and not expiration: expiration = int(v)
                if k == 'serializer': serializer = v
                if k == 'async_enabled': async_enabled = v in {'true', '1', 't', 'y', 'yes'}
                if k == 'auto_delete_invalid': auto_delete_invalid = v in {'true', '1', 't', 'y', 'yes'}
        
        # Now we'll configure the rest
        if name is not None: self.name = name
        if table is not None: self.table = table
        if expiration is not None:  self.expiration = expiration
        if optimization is not None: self.optimization = optimization
        self.async_enabled = async_enabled
        if auto_delete_invalid is not None: self.auto_delete_invalid = auto_delete_invalid
        from lzl.io.ser import get_serializer
        serializer_kwargs = serializer_kwargs or {}
        if 'verbosity' not in serializer_kwargs:
            serializer_kwargs['verbosity'] = 0
        self.serializer = get_serializer(serializer = serializer, **serializer_kwargs)
        self.db = SqliteDB(conn_uri = conn_uri, table = self.table, optimization = self.optimization, configured = configured, is_remote = is_remote, **(db_settings or {})) if \
            _db is None else _db
    
        self._kwargs = kwargs
        self._kwargs['serializer'] = serializer
        self._kwargs['serializer_kwargs'] = serializer_kwargs
        self._kwargs['async_enabled'] = async_enabled
        self._kwargs['expiration'] = expiration
        self._kwargs['name'] = name
        self._kwargs['optimization'] = optimization
        self._kwargs['conn_uri'] = conn_uri
        self._kwargs['db_settings'] = db_settings
        self._kwargs['is_remote'] = is_remote
        self._kwargs['auto_delete_invalid'] = auto_delete_invalid


    def getter(self, key: str, default: Optional[Any] = None, is_async: Optional[bool] = None) -> t.Union[Callable[[Union[str, bytes]], Any], Callable[..., t.Awaitable[Union[str, bytes]]]]:
        """
        Returns a getter function
        """
        return self.db.aget(key, default = default) if is_async else self.db.get(key, default = default)

    def setter(self, key: str, value: t.Union[str, bytes], ex: Optional[int] = None, tag: t.Optional[str] = None, is_async: Optional[bool] = None) -> t.Union[Callable[[Union[str, bytes], Any], Any], Callable[..., t.Awaitable[Any]]]:
        """
        Returns a setter function
        """
        ex = ex or self.expiration
        return self.db.aset(key, value, expire = ex, tag = tag) if is_async else self.db.set(key, value, expire = ex, tag = tag)

    def deleter(self, key: str, is_async: Optional[bool] = None):
        """
        Returns a deleter function
        """
        return self.db.adelete(key) if is_async else self.db.delete(key)
    
    def deleters(self, *keys: str, is_async: Optional[bool] = None) -> int:
        """
        Returns a deleter function
        """
        return self.db._delete_keys_(*keys) if is_async else self.db._adelete_keys_(*keys)
    
    def _contains(self, key: str, is_async: Optional[bool] = None) -> bool:
        """
        Returns whether the key is in the cache
        """
        return self.db.acontains(key) if is_async else self.db.contains(key)

    def _precheck(self, **kwargs):
        """
        Run a precheck operation
        """
        pass

    def get_child(self, key: str, **kwargs) -> 'SqliteStatefulBackend':
        """
        Gets a Child Persistent Dictionary
        """
        base_kwargs = copy.deepcopy(self._kwargs)
        base_kwargs.update(kwargs)
        # base_kwargs['db_settings'] = base_kwargs.get('db_settings', {}) or {}
        # logger.info(base_kwargs)
        db = self.db.get_child(key, timeout = base_kwargs.get('timeout'), **(base_kwargs.get('db_settings') or {}))
        return self.__class__(
            _db = db, 
            **base_kwargs
        )


    """
    Helper Methods
    """

    def _get_one(
        self, 
        key: str, 
        default: Optional[Any] = None, 
        _raw: Optional[bool] = None, 
        **kwargs
    ) -> Optional[Any]:
        """
        Fetches a Value from the DB
        """
        value = self.getter(key, default = ENOVAL, is_async = False)
        if value is ENOVAL: return default
        if _raw: return value
        try:
            decoded = self.decode_value(value, _raw = _raw, **kwargs)
            return default if decoded is None else decoded
        except Exception as e:
            logger.info(f'Error Decoding Value: |r|({type(value)}) {e}|e| {value}', colored = True)
            if self.auto_delete_invalid: self.deleter(key)
            return default

    async def _aget_one(
        self, 
        key: str, 
        default: Optional[Any] = None, 
        _raw: Optional[bool] = None, 
        **kwargs
    ) -> Optional[Any]:
        """
        [Async] Fetches a Value from the DB
        """
        value = await self.getter(key, default = ENOVAL, is_async = True)
        if value is ENOVAL: return default
        if _raw: return value
        try:
            decoded = await self.adecode_value(value, _raw = _raw, **kwargs)
            return default if decoded is None else decoded
        except Exception as e:
            logger.info(f'Error Decoding Value: |r|({type(value)}) {e}|e| {value}', colored = True)
            if self.auto_delete_invalid: await self.deleter(key)
            return default
        
    def _set_one(
        self, 
        key: str, 
        value: Any, 
        ex: Optional[int] = None, 
        tag: Optional[str] = None, 
        _raw: Optional[bool] = None, 
        **kwargs
    ):
        """
        Saves a Value to the DB
        """
        try:
            value = self.encode_value(value, _raw = _raw, **kwargs)
        except Exception as e:
            logger.info(f'Error Encoding Value: |r|({type(value)}) {e}|e| {value}', colored = True)
            return None
        try:
            return self.setter(key, value, ex = ex, tag = tag, is_async = False)
        except Exception as e:
            logger.info(f'Error Saving Value: |r|({type(value)}) {e}|e| {value}', colored = True)
            return None
        
    async def _aset_one(
        self, 
        key: str, 
        value: Any, 
        ex: Optional[int] = None, 
        tag: Optional[str] = None, 
        _raw: Optional[bool] = None, 
        **kwargs
    ):
        """
        [Async] Saves a Value to the DB
        """
        try:
            value = await self.aencode_value(value, _raw = _raw, **kwargs)
        except Exception as e:
            logger.info(f'Error Encoding Value: |r|({type(value)}) {e}|e| {value}', colored = True)
            return None
        try:
            return await self.setter(key, value, ex = ex, tag = tag, is_async = True)
        except Exception as e:
            logger.info(f'Error Saving Value: |r|({type(value)}) {e}|e| {value}', colored = True)
            return None


    """
    Encoding and Decoding Methods
    """

    # def encode_value(self, value: t.Union[t.Any, SchemaType], _raw: t.Optional[bool] = None, **kwargs) -> t.Union[str, bytes]:
    #     """
    #     Encodes a Value
    #     """
    #     return value if _raw else self.serializer.encode(value, **kwargs)
    
    # async def aencode_value(self, value: t.Union[t.Any, SchemaType], _raw: Optional[bool] = None,  **kwargs) -> t.Union[str, bytes]:
    #     """
    #     Encodes a Value
    #     """
    #     return value if _raw else await self.serializer.aencode(value, **kwargs)

    # def decode_value(self, value: t.Union[str, bytes], _raw: t.Optional[bool] = None, **kwargs) -> t.Any:
    #     """
    #     Decodes a Value
    #     """
    #     return value if _raw else self.serializer.decode(value, **kwargs)
    
    # async def adecode_value(self, value: t.Union[str, bytes], _raw: t.Optional[bool] = None, **kwargs) -> t.Any:
    #     """
    #     Decodes a Value
    #     """
    #     return value if _raw else await self.serializer.adecode(value, **kwargs)
    

    def _decode_kv_data(self, data: t.Dict[str, t.Any], _raw: Optional[bool] = None, skip_on_errors: Optional[bool] = True, **kwargs) -> t.Dict[str, t.Any]:
        """
        Decodes the data
        """
        data_keys = list(data.keys())
        for key in data_keys:
            try:
                data[key] = self.decode_value(data[key], _raw = _raw, **kwargs)
            except Exception as e:
                if skip_on_errors: 
                    data.pop(key)
                    continue
                logger.info(f'Error Decoding Key {key} : |r|({type(data[key])}) {e}|e| {data[key]}', colored = True)
                raise e
        return data

    async def _adecode_kv_data(self, data: t.Dict[str, t.Any], _raw: Optional[bool] = None, skip_on_errors: Optional[bool] = True, **kwargs) -> t.Dict[str, t.Any]:
        """
        [Async] Decodes the data
        """
        data_keys = list(data.keys())
        for key in data_keys:
            try:
                data[key] = await self.adecode_value(data[key], _raw = _raw, **kwargs)
            except Exception as e:
                if skip_on_errors: 
                    data.pop(key)
                    continue
                logger.info(f'Error Decoding Key {key} : |r|({type(data[key])}) {e}|e| {data[key]}', colored = True)
                raise e
        return data
    
    def _decode_values(self, values: t.List[t.Any], _raw: Optional[bool] = None, skip_on_errors: Optional[bool] = True, **kwargs) -> t.List[t.Any]:
        """
        Decodes the values
        """
        invalid_idx = []
        for n, value in enumerate(values):
            try:
                values[n] = self.decode_value(value, _raw = _raw, **kwargs)
            except Exception as e:
                if skip_on_errors: 
                    invalid_idx.append(n)
                    # values.pop(n)
                    continue
                logger.info(f'Error Decoding Value {n} : |r|({type(value)}) {e}|e| {value}', colored = True)
                raise e
        if invalid_idx:
            values = [values[i] for i in range(len(values)) if i not in invalid_idx]
        return values

    async def _adecode_values(self, values: t.List[t.Any], _raw: Optional[bool] = None, skip_on_errors: Optional[bool] = True, **kwargs) -> t.List[t.Any]:
        """
        [Async] Decodes the values
        """
        invalid_idx = []
        for n, value in enumerate(values):
            try:
                values[n] = await self.adecode_value(value, _raw = _raw, **kwargs)
            except Exception as e:
                if skip_on_errors: 
                    invalid_idx.append(n)
                    # values.pop(n)
                    continue
                logger.info(f'Error Decoding Value {n} : |r|({type(value)}) {e}|e| {value}', colored = True)
                raise e
        if invalid_idx:
            values = [values[i] for i in range(len(values)) if i not in invalid_idx]
        return values

    """
    Implemented Methods
    """

    def get_key(self, key: str) -> str:
        """
        Gets a Key
        """
        # Since we use a table, we just return the key
        return key
    
    def get(self, key: str, default: Optional[Any] = None, _raw: Optional[bool] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from SQLite
        """
        return self._get_one(key, default = default, _raw = _raw, **kwargs)
    
    async def aget(self, key: str, default: Optional[Any] = None, _raw: Optional[bool] = None, **kwargs) -> Optional[Any]:
        """
        [Async] Gets a Value from SQLite
        """
        return await self._aget_one(key, default = default, _raw = _raw, **kwargs)
    
    def set(self, key: str, value: Any, ex: Optional[int] = None, tag: Optional[str] = None, _raw: Optional[bool] = None, **kwargs) -> None:
        """
        Sets a Value in SQLite
        """
        return self._set_one(key, value, ex = ex, tag = tag, _raw = _raw, **kwargs)
    
    async def aset(self, key: str, value: Any, ex: Optional[int] = None, tag: Optional[str] = None, _raw: Optional[bool] = None, **kwargs) -> None:
        """
        [Async] Sets a Value in SQLite
        """
        return await self._aset_one(key, value, ex = ex, tag = tag, _raw = _raw, **kwargs)
    

    def set_batch(self, data: Dict[str, Any], ex: Optional[int] = None, tag: t.Optional[str] = None, _raw: Optional[bool] = None, skip_on_errors: t.Optional[bool] = True, **kwargs) -> int:
        """
        Saves a Value to the Object Store
        """
        data_keys = list(data.keys())
        for key in data_keys:
            try:
                data[key] = self.encode_value(data[key], _raw = _raw, **kwargs)
            except Exception as e:
                value = data.pop(key)
                logger.info(f'Error Encoding Key {key} : |r|({type(value)}) {e}|e| {value}', colored = True)
                if skip_on_errors: continue
                raise e
        ex = ex or self.expiration
        self.db.batch_set(data, expire = ex, tag = tag, **kwargs)
        return len(data)

    async def aset_batch(self, data: Dict[str, Any], ex: Optional[int] = None, tag: t.Optional[str] = None, _raw: Optional[bool] = None, skip_on_errors: t.Optional[bool] = True, **kwargs) -> int:
        """
        [Async] Saves a Value to the Object Store
        """
        data_keys = list(data.keys())
        for key in data_keys:
            try:
                data[key] = await self.aencode_value(data[key], _raw = _raw, **kwargs)
            except Exception as e:
                value = data.pop(key)
                logger.info(f'Error Encoding Key {key} : |r|({type(value)}) {e}|e| {value}', colored = True)
                if skip_on_errors: continue
                raise e
        ex = ex or self.expiration
        await self.db.abatch_set(data, expire = ex, tag = tag, **kwargs)
        return len(data)


        
    def delete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from SQLite
        """
        return self.deleter(key, **kwargs)
    
    async def adelete(self, key: str, **kwargs) -> None:
        """
        [Async] Deletes a Value from SQLite
        """
        return await self.deleters(key, is_async = True, **kwargs)

    def get_keys(
        self, 
        pattern: str, 
        exclude_base_key: Optional[bool] = None, 
        order: Optional[str] = 'ASC',
        limit: Optional[int] = None,
        **kwargs
    ) -> List[str]:
        """
        Returns the Keys within the DB
        """
        if '*' in pattern: pattern = pattern.replace('*', '%')
        return self.db.fetch_keys(pattern = pattern, include_rowid = False, order = order, limit = limit, **kwargs)
    
    async def aget_keys(
        self, 
        pattern: str, 
        exclude_base_key: Optional[bool] = None, 
        order: Optional[str] = 'ASC',
        limit: Optional[int] = None,
        **kwargs
    ) -> List[str]:
        """
        [Async] Returns the Keys within the DB
        """
        if '*' in pattern: pattern = pattern.replace('*', '%')
        return await self.db.afetch_keys(pattern = pattern, include_rowid = False, order = order, limit = limit, **kwargs)
    
    def get_all_keys(
        self, 
        *args,
        **kwargs
    ) -> List[str]:
        """
        Returns all the Keys in the DB
        """
        return self.db.fetch_keys(pattern = None, **kwargs)

    async def aget_all_keys(
        self, 
        *args,
        **kwargs
    ) -> List[str]:
        """
        [Async] Returns all the Keys in the DB
        """
        return await self.db.afetch_keys(pattern = None, **kwargs)

    def get_all_data(self, exclude_base_key: Optional[bool] = False, limit: t.Optional[int] = None, _raw: Optional[bool] = None, skip_on_errors: t.Optional[bool] = True, **kwargs) -> Dict[str, t.Any]:
        """
        Loads all the Data
        """
        
        data = self.db.fetch_kv_data(limit = limit, **kwargs)
        return self._decode_kv_data(data, _raw = _raw, skip_on_errors = skip_on_errors, **kwargs)
        

    async def aget_all_data(self, exclude_base_key: Optional[bool] = False, limit: t.Optional[int] = None, _raw: Optional[bool] = None, skip_on_errors: t.Optional[bool] = True, **kwargs) -> Dict[str, t.Any]:
        """
        [Async] Loads all the Data
        """
        data = await self.db.afetch_kv_data(limit = limit, **kwargs)
        return await self._adecode_kv_data(data, _raw = _raw, skip_on_errors = skip_on_errors, **kwargs)


    def get_all_values(self, limit: t.Optional[int] = None, _raw: Optional[bool] = None, skip_on_errors: t.Optional[bool] = True, **kwargs) -> List[t.Any]:
        """
        Returns all the Values in the DB
        """
        values = self.db.fetch_values(limit = limit, **kwargs)
        return self._decode_values(values, _raw = _raw, skip_on_errors = skip_on_errors, **kwargs)
        
    async def aget_all_values(self, limit: t.Optional[int] = None, _raw: Optional[bool] = None, skip_on_errors: t.Optional[bool] = True, **kwargs) -> List[t.Any]:
        """
        [Async] Returns all the Values in the DB
        """
        values = await self.db.afetch_values(limit = limit, **kwargs)
        return await self._adecode_values(values, _raw = _raw, skip_on_errors = skip_on_errors, **kwargs)


    def length(self, **kwargs) -> int:
        """
        Returns the Length of the DB
        """
        return self.db.length(**kwargs)
    
    async def alength(self, **kwargs) -> int:
        """
        [Async] Returns the Length of the DB
        """
        return await self.db.alength(**kwargs)
    

    def __len__(self):
        """
        Returns the Length of the Cache
        """
        return self.length()
    
    def clear(self, **kwargs) -> None:
        """
        Clears the DB
        """
        return self.db.clear(**kwargs)
    
    async def aclear(self, **kwargs) -> None:
        """
        [Async] Clears the DB
        """
        return await self.db.aclear(**kwargs)

    # def iterate(self, **kwargs) -> t.Iterable[t.Any]:
    #     """
    #     Iterates over the Cache
    #     """
    #     return self.db.iterate(**kwargs) 
    
    def contains(self, key: str, **kwargs) -> bool:
        """
        Returns whether the Key is in the DB
        """
        return self.db.contains(key, **kwargs)
    
    async def acontains(self, key: str, **kwargs) -> bool:
        """
        [Async] Returns whether the Key is in the DB
        """
        return await self.db.acontains(key, **kwargs)
    
    def expire(self, key: str, ex: int, **kwargs) -> None:
        """
        Expires a Key
        """
        return self.db._set_key_expiration_(key, ex, **kwargs)
    
    async def aexpire(self, key: str, ex: int, **kwargs) -> None:
        """
        [Async] Expires a Key
        """
        return await self.db._aset_key_expiration_(key, ex, **kwargs)

    def update(self, data: t.Dict[str, t.Any], ex: t.Optional[int] = None, tag: t.Optional[str] = None, retry: t.Optional[bool] = None, **kwargs):
        """
        Updates the Cache
        """
        return self.set_batch(data, ex = ex, tag = tag, retry = retry)

    async def aupdate(self, data: t.Dict[str, t.Any], ex: t.Optional[int] = None, tag: t.Optional[str] = None, retry: t.Optional[bool] = None, **kwargs):
        """
        [Async] Updates the Cache
        """
        return await self.aset_batch(data, ex = ex, tag = tag, retry = retry)

    def select(
        self, 
        *tags: str, 
        limit: t.Optional[int] = None, 
        order: t.Optional[t.Literal['ASC', 'DESC']] = 'ASC',
        order_by: t.Optional[str] = 'rowid',
        include_meta: bool = False, 
        _raw: t.Optional[bool] = None,
        skip_on_errors: t.Optional[bool] = True,
        **kwargs
    ) -> t.Union[t.Dict[str, t.Any], t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]]:
        """
        Select all items with the given tags

        :param tags: tags to select
        :param limit: limit the number of items to return
            (default None, no limit)
        :param order: order to return the items
            (default 'ASC', ascending)
        :param include_meta: if True, return a dict of metadata
            (default False)
        :return: dict of key/value pairs or tuple of (key, value) pairs if include_meta is True
        """
        data = self.db.select_tags(*tags, limit = limit, order = order, order_by = order_by, include_meta = include_meta, **kwargs)
        if include_meta: 
            data, meta = data
        data = self._decode_kv_data(data, _raw = _raw, skip_on_errors = skip_on_errors, **kwargs)
        return (data, meta) if include_meta else data
    
    
    async def aselect(
        self, 
        *tags: str, 
        limit: t.Optional[int] = None, 
        order: t.Optional[t.Literal['ASC', 'DESC']] = 'ASC',
        order_by: t.Optional[str] = 'rowid',
        include_meta: bool = False, 
        _raw: t.Optional[bool] = None,
        skip_on_errors: t.Optional[bool] = True,
        **kwargs
    ) -> t.Union[t.Dict[str, t.Any], t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]]:
        """
        [Async] Select all items with the given tags

        :param tags: tags to select
        :param limit: limit the number of items to return
            (default None, no limit)
        :param order: order to return the items
            (default 'ASC', ascending)
        :param include_meta: if True, return a dict of metadata
            (default False)
        :return: dict of key/value pairs or tuple of (key, value) pairs if include_meta is True
        """
        data = await self.db.aselect_tags(*tags, limit = limit, order = order, order_by = order_by, include_meta = include_meta, **kwargs)
        if include_meta: 
            data, meta = data
        data = await self._adecode_kv_data(data, _raw = _raw, skip_on_errors = skip_on_errors, **kwargs)
        return (data, meta) if include_meta else data

    def __repr__(self):
        """
        Returns the Representation of the Cache
        """
        return f"<{self.__class__.__name__} num_keys={len(self)}, table={self.db.table}, serializer={self.serializer.name}>"




DEFAULT_STATUS_COLORS = {
    'debug': '<fg #D9ED92>',
    'info': '<fg #34A0A4>',
    'success': '<fg #52B69A>',
    'warning': '<fg #F48C06>',
    'error': '<fg #DC2F02>',
    'critical': '<fg #9D0208>',
}
    

QUEUE_STATUS_COLORS = {
    'new': '<fg #D9ED92>',
    'deferred': '<fg #B5E48C>',
    'queued': '<fg #99D98C>',
    'active': '<fg #76C893>',
    'complete': '<fg #52B69A>',

    # Error Colors
    'aborted': '<fg #FFBA08>',
    'failed': '<fg #9D0208>',

    # Other Colors
    'enqueue': '<fg #168AAD>',
    'finish': '<fg #52B69A>',
    'completed': '<fg #52B69A>',
    'error': '<fg #DC2F02>',
    'abort': '<fg #DC2F02>',

    'retry': '<fg #F48C06>',
    'scheduled': '<fg #34A0A4>',
    'reschedule': '<fg #34A0A4>',
    'startup': '<fg #168AAD>',
    'shutdown': '<fg #6A040F>',
    'process': '<fg #184E77>',
    'sweep': '<fg #B5E48C>',
    'stats': '<fg #B5E48C>',
    'dequeue': '<fg #168AAD>',

    'stuck': '<fg #DC2F02>',
}

STATUS_COLOR = QUEUE_STATUS_COLORS
FALLBACK_STATUS_COLOR = '<fg #99D98C>'

# DEFAULT_FUNCTION_COLOR = '<fg #457b9d>'
DEFAULT_FUNCTION_COLOR = '<fg #219ebc>'
DEFAULT_CLASS_COLOR = '<fg #a8dadc>'
RESET_COLOR = '\x1b[0m'

LOGLEVEL_MAPPING = {
    50: 'CRITICAL',
    40: 'ERROR',
    30: 'WARNING',
    20: 'INFO',
    19: 'DEV',
    10: 'DEBUG',
    5: 'CRITICAL',
    4: 'ERROR',
    3: 'WARNING',
    2: 'INFO',
    1: 'DEBUG',
    0: 'NOTSET',
}

REVERSE_LOGLEVEL_MAPPING = {v: k for k, v in LOGLEVEL_MAPPING.items()}
COLORED_MESSAGE_MAP = {
    '|bld|': '<bold>',
    '|reset|': '</>',
    '|eee|': '</></></>',
    '|em|': '<bold>',
    '|ee|': '</></>',
    '|lr|': '<light-red>',
    '|lb|': '<light-blue>',
    '|lm|': '<light-magenta>',
    '|lc|': '<light-cyan>',
    '|lw|': '<light-white>',
    # '|gr|': '<gray>',
    # '|gr|': "\033[90m",
    '|gr|': '<fg #808080>',
    '|lk|': '<light-black>',
    '|br|': "\x1b[31;1m", # Bold Red
    '|k|': '<black>',
    '|r|': '<red>',
    '|m|': '<magenta>',
    '|c|': '<cyan>',
    '|u|': '<underline>',
    '|i|': '<italic>',
    '|s|': '<strike>',
    '|e|': '</>',
    '|g|': '<green>',
    '|y|': '<yellow>',
    '|b|': '<blue>',
    '|w|': '<white>',
}



from __future__ import annotations

"""
File Decoders
"""

# from lzl import load

try:
    from httpx._decoders import ByteChunker, TextDecoder, TextChunker, LineDecoder

except ImportError:

    import io
    import codecs

    class ByteChunker:
        """
        Handles returning byte content in fixed-size chunks.
        """

        def __init__(self, chunk_size: int | None = None) -> None:
            self._buffer = io.BytesIO()
            self._chunk_size = chunk_size

        def decode(self, content: bytes) -> list[bytes]:
            if self._chunk_size is None:
                return [content] if content else []

            self._buffer.write(content)
            if self._buffer.tell() < self._chunk_size:
                return []
            value = self._buffer.getvalue()
            chunks = [
                value[i : i + self._chunk_size]
                for i in range(0, len(value), self._chunk_size)
            ]
            if len(chunks[-1]) == self._chunk_size:
                self._buffer.seek(0)
                self._buffer.truncate()
                return chunks
            else:
                self._buffer.seek(0)
                self._buffer.write(chunks[-1])
                self._buffer.truncate()
                return chunks[:-1]

        def flush(self) -> list[bytes]:
            value = self._buffer.getvalue()
            self._buffer.seek(0)
            self._buffer.truncate()
            return [value] if value else []
        

    class TextChunker:
        """
        Handles returning text content in fixed-size chunks.
        """

        def __init__(self, chunk_size: int | None = None) -> None:
            self._buffer = io.StringIO()
            self._chunk_size = chunk_size

        def decode(self, content: str) -> list[str]:
            if self._chunk_size is None:
                return [content] if content else []

            self._buffer.write(content)
            if self._buffer.tell() < self._chunk_size:
                return []
            value = self._buffer.getvalue()
            chunks = [
                value[i : i + self._chunk_size]
                for i in range(0, len(value), self._chunk_size)
            ]
            if len(chunks[-1]) == self._chunk_size:
                self._buffer.seek(0)
                self._buffer.truncate()
                return chunks
            else:
                self._buffer.seek(0)
                self._buffer.write(chunks[-1])
                self._buffer.truncate()
                return chunks[:-1]

        def flush(self) -> list[str]:
            value = self._buffer.getvalue()
            self._buffer.seek(0)
            self._buffer.truncate()
            return [value] if value else []


    class TextDecoder:
        """
        Handles incrementally decoding bytes into text
        """

        def __init__(self, encoding: str = "utf-8") -> None:
            self.decoder = codecs.getincrementaldecoder(encoding)(errors="replace")

        def decode(self, data: bytes) -> str:
            return self.decoder.decode(data)

        def flush(self) -> str:
            return self.decoder.decode(b"", True)


    class LineDecoder:
        """
        Handles incrementally reading lines from text.

        Has the same behaviour as the stdllib splitlines,
        but handling the input iteratively.
        """

        def __init__(self) -> None:
            self.buffer: list[str] = []
            self.trailing_cr: bool = False

        def decode(self, text: str) -> list[str]:
            # See https://docs.python.org/3/library/stdtypes.html#str.splitlines
            NEWLINE_CHARS = "\n\r\x0b\x0c\x1c\x1d\x1e\x85\u2028\u2029"

            # We always push a trailing `\r` into the next decode iteration.
            if self.trailing_cr:
                text = "\r" + text
                self.trailing_cr = False
            if text.endswith("\r"):
                self.trailing_cr = True
                text = text[:-1]

            if not text:
                # NOTE: the edge case input of empty text doesn't occur in practice,
                # because other httpx internals filter out this value
                return []  # pragma: no cover

            trailing_newline = text[-1] in NEWLINE_CHARS
            lines = text.splitlines()

            if len(lines) == 1 and not trailing_newline:
                # No new lines, buffer the input and continue.
                self.buffer.append(lines[0])
                return []

            if self.buffer:
                # Include any existing buffer in the first portion of the
                # splitlines result.
                lines = ["".join(self.buffer) + lines[0]] + lines[1:]
                self.buffer = []

            if not trailing_newline:
                # If the last segment of splitlines is not newline terminated,
                # then drop it from our output and start a new buffer.
                self.buffer = [lines.pop()]

            return lines

        def flush(self) -> list[str]:
            if not self.buffer and not self.trailing_cr:
                return []

            lines = ["".join(self.buffer)]
            self.buffer = []
            self.trailing_cr = False
            return lines


"""Logging helpers specific to the :mod:`lzl.api.aiohttpx` package."""

import logging
import os
import typing as t

from lzl.logging import Logger, NullLogger, change_logger_level, get_logger, null_logger

__all__ = [
    "logger",
    "mute_httpx_logger",
    "change_logger_level",
    "null_logger",
    "Logger",
    "NullLogger",
]

# Keep module-level logger creation simple to avoid recursive import issues.
logger_level = os.getenv('LOGGER_LEVEL', 'INFO').upper()
logger: Logger = get_logger(logger_level)
logger.set_module_name('lzl.api.aiohttpx', 'aiohttpx', is_relative=True)

_muted_httpx: t.Optional[bool] = None


def mute_httpx_logger() -> None:
    """Ensure the upstream ``httpx`` logger only logs warnings or higher once."""

    global _muted_httpx
    if _muted_httpx is not None:
        return
    _muted_httpx = True
    logging.getLogger("httpx").setLevel(logging.WARNING)



from __future__ import annotations

"""Convenience registries that map adapters to backend implementations."""

import typing as t

from .base.backend import BaseDatabaseBackend, BackendClassT
from .base.manager import DBBackendManager, DBManagerT
from .postgres import PostgresBackendManager, PostgresBackendMap, PostgresBackendClasses
from .sqlite import SqliteBackendManager, SqliteBackendMap, SqliteBackendClasses

ADAPTER_TO_BACKENDS: t.Dict[str, t.Dict[str, t.Type[BackendClassT]]] = {
    'postgres': PostgresBackendMap,
    'postgresql': PostgresBackendMap,
    'sqlite': SqliteBackendMap,
}

ADAPTER_TO_MANAGER: t.Dict[str, t.Type[DBManagerT]] = {
    'postgres': PostgresBackendManager,
    'postgresql': PostgresBackendManager,
    'sqlite': SqliteBackendManager,
}

DATABASE_MANAGER_CLASSES = t.Union[DBBackendManager, SqliteBackendManager, PostgresBackendManager]
DATABASE_BACKEND_CLASSES = t.Union[BaseDatabaseBackend, SqliteBackendClasses, PostgresBackendClasses]

__all__ = [
    "ADAPTER_TO_BACKENDS",
    "ADAPTER_TO_MANAGER",
    "DATABASE_MANAGER_CLASSES",
    "DATABASE_BACKEND_CLASSES",
]



from __future__ import annotations

import json
import logging
import pathlib
import contextlib
from lzl.api import aiohttpx
from pydantic import PrivateAttr
from lzl.types import BaseSettings, validator, eproperty, Field, field_validator
from lzl.proxied import proxied, ProxyObject
from lzo.types import AppEnv
from lzl.api.openai.version import VERSION, DEFAULT_AZURE_VERSION
from lzl.api.openai.types.auth import OpenAIAuth, OpenAIContext
from lzl.api.openai.types.base import ApiType
from lzl.api.openai.configs.proxy import OpenAIProxySettings
from typing import Optional, Dict, Union, Any



class BaseOpenAISettings(BaseSettings):
    url: Optional[str] = None
    scheme: Optional[str] = 'http://'
    host: Optional[str] = None
    port: Optional[int] = 8080
    
    api_base: Optional[str] = None
    api_key: Optional[str] = None
    api_path: Optional[str] = '/v1'
    api_type: Optional[ApiType] = ApiType.openai
    api_version: Optional[str] = None
    api_key_path: Optional[pathlib.Path] = None
    
    organization: Optional[str] = None
    proxies: Optional[Union[Dict, str]] = None
    app_info: Optional[Dict[str, str]] = None
    debug_enabled: Optional[bool] = False
    ignore_errors: Optional[bool] = False
    disable_retries: Optional[bool] = False # Allows users to customize the retry behavior

    timeout: Optional[int] = 600
    max_retries: Optional[int] = 3

    # Additional Configuration
    ## Request Pool Configuration
    max_connections: Optional[int] = 250
    max_keepalive_connections: Optional[int] = 150
    keepalive_expiry: Optional[int] = 60

    custom_headers: Optional[Dict[str, str]] = None
    limit_monitor_enabled: Optional[bool] = True

    _extra: Dict[str, Any] = PrivateAttr(default_factory = dict)

    @validator("api_type")
    def validate_api_type(cls, v):
        """
        Validates the API Type
        """
        if v is None: return ApiType.openai
        return ApiType.parse(v) if isinstance(v, str) else v
    
    @validator("api_key")
    def validate_api_key(cls, v, values: Dict[str, Union[str, int, bool, pathlib.Path, Any]]):
        """
        Validates the API Key
        """
        if v is None and values.get('api_key_path') is not None:
            return values['api_key_path'].read_text()
        return v
    
    @eproperty
    def ctx(self) -> OpenAIContext:
        """
        Returns the context
        """
        return OpenAIContext(
            custom_headers = self.custom_headers or {},
        )

    @eproperty
    def api_url(self) -> str:
        """
        Returns the API URL
        """
        if self.api_base: return self.api_base
        if self.url: return self.url
        if self.host:
            url = f"{self.scheme}{self.host}"
            if self.port: url += f":{self.port}"
            return url
        
        # Return the official Open API URL
        return "https://api.openai.com"
    
    @eproperty
    def api_client_limits(self) -> aiohttpx.Limits:
        """
        Returns the API Client Limits
        """
        return aiohttpx.Limits(
            max_connections = self.max_connections,
            max_keepalive_connections = self.max_keepalive_connections,
            keepalive_expiry = self.keepalive_expiry,
        )
    

    @eproperty
    def base_url(self) -> str:
        """
        Returns the Base URL
        """
        if self.api_path:
            from urllib.parse import urljoin
            return urljoin(self.api_url, self.api_path)
        return self.api_url
    
    @eproperty
    def base_headers(self) -> Dict[str, str]:
        """
        Returns the Base Headers
        """
        if 'app_headers' not in self.ctx.data:
            import platform
            ua = f"OpenAI/v1 async_openai/{VERSION}"
            if self.app_info:
                t = ""
                if "name" in self.app_info:
                    t += self.app_info["name"]
                if "version" in self.app_info:
                    t += f"/{self.app_info['version']}"
                if "url" in self.app_info:
                    t += f" ({self.app_info['url']})"
                ua += f" {t}"
            uname_without_node = " ".join(
                v for k, v in platform.uname()._asdict().items() if k != "node"
            )
            data = {
                "bindings_version": VERSION,
                "httplib": "httpx",
                "lang": "python",
                "lang_version": platform.python_version(),
                "platform": platform.platform(),
                "publisher": "trisongz",
                "uname": uname_without_node,
            }
            if self.app_info: data["application"] = self.app_info
            self.ctx.data['app_headers'] = {"X-OpenAI-Client-User-Agent": json.dumps(data), "User-Agent": ua}
        return self.ctx.data['app_headers']
    
    # Deprecated
    def get_api_key_headers(
        self,
        api_key: Optional[str] = None, 
        api_type: Optional[Union[ApiType, str]] = None
    ) -> Dict[str, str]:
        """
        Returns the API Key Headers
        """
        if api_key is None: api_key = self.api_key
        if api_type is None: api_type = self.api_type
        api_type = api_type.value if isinstance(api_type, ApiType) else api_type
        if api_type in {"openai", "azure_ad"}:
            return {"Authorization": f"Bearer {api_key}"}
        return {"api-key": api_key}

    def get_api_client_auth(
        self,
        api_key: Optional[str] = None, 
        api_type: Optional[Union[ApiType, str]] = None,
        **kwargs
    ) -> OpenAIAuth:
        """
        Returns the API Client Auth
        """
        if api_key is None: api_key = self.api_key
        if api_type is None: api_type = self.api_type
        api_type = api_type.value if isinstance(api_type, ApiType) else api_type
        if api_type in {"openai", "azure_ad"}:
            return OpenAIAuth(
                settings = self, context = self.ctx, auth_key = "Authorization", auth_value = f"Bearer {api_key}",
            )
        return OpenAIAuth(
            settings = self, context = self.ctx, auth_key = "api-key", auth_value = api_key,
        )

    def get_headers(
        self, 
        # api_key: Optional[str] = None, 
        api_version: Optional[str] = None,
        api_type: Optional[Union[ApiType, str]] = None,
        organization: Optional[str] = None,
        app_info: Optional[str] = None,
        **kwargs
    ) -> Dict[str, str]:
        """
        Gets the Headers
        """
        # print(api_key, api_version, api_type, organization, kwargs)
        # headers = self.headers.copy()
        headers = self.base_headers.copy()
        if kwargs: headers.update(**kwargs)
        if app_info is not None: headers['application'] = app_info
        # headers.update(self.get_api_key_headers(api_key = api_key, api_type = api_type))
        if organization is None: organization = self.organization
        if api_version is None: api_version = self.api_version
        if organization:
            headers["OpenAI-Organization"] = organization
        if api_version is not None and api_type.value == 'open_ai':
            headers["OpenAI-Version"] = api_version
        if self.debug_enabled:
            headers["OpenAI-Debug"] = "true"
        headers['Content-Type'] = 'application/json'
        return headers

    def get_api_url(
        self, 
        host: Optional[str] = None, 
        port: Optional[int] = None, 
        scheme: Optional[str] = None, 
        url: Optional[str] = None,
        api_base: Optional[str] = None,
        **kwargs
    ) -> str:
        """
        Returns the API URL
        """
        if api_base: return api_base
        if url: return url
        if host:
            url = f"{scheme or self.scheme}{host}"
            if port: url += f":{port}"
            return url
        return self.api_url

    def get_base_api_url(
        self, 
        host: Optional[str] = None, 
        port: Optional[int] = None, 
        scheme: Optional[str] = None, 
        url: Optional[str] = None,
        api_path: Optional[str] = None,
        api_base: Optional[str] = None,
        **kwargs
    ) -> str:
        """
        Returns the Base API URL
        """
        api_url = self.get_api_url(
            host=host,
            port=port,
            scheme=scheme,
            url=url,
            api_base=api_base,
        )
        api_path = api_path or self.api_path
        if api_path:
            from urllib.parse import urljoin
            return urljoin(api_url, api_path)
        return api_url


    def configure(
        self, 
        url: Optional[str] = None,
        scheme: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        
        api_base: Optional[str] = None,
        api_key: Optional[str] = None,
        api_path: Optional[str] = None,
        api_type: Optional[ApiType] = None,
        api_version: Optional[str] = None,
        api_key_path: Optional[pathlib.Path] = None,

        organization: Optional[str] = None,
        proxies: Optional[Union[str, Dict]] = None,
        timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
        app_info: Optional[Dict[str, str]] = None,
        debug_enabled: Optional[bool] = None,
        ignore_errors: Optional[bool] = None,
        disable_retries: Optional[bool] = None,

        max_connections: Optional[int] = None,
        max_keepalive_connections: Optional[int] = None,
        keepalive_expiry: Optional[int] = None,
        custom_headers: Optional[Dict[str, str]] = None,

        **kwargs
    ):  # sourcery skip: low-code-quality
        """
        Allows Post-Init Configuration of the OpenAI Settings
        """
        if url is not None: self.url = url
        if api_base is not None: self.api_base = api_base
        if scheme is not None: self.scheme = scheme
        if host is not None: self.host = host
        if port is not None: self.port = port
        if api_key is not None: self.api_key = api_key
        if api_path is not None: self.api_path = api_path
        if api_key_path is not None: 
            self.api_key_path = api_key_path if isinstance(api_key_path, pathlib.Path) else pathlib.Path(api_key_path)
            self.api_key = self.api_key_path.read_text().strip()
        
        if api_type is not None: 
            self.api_type = ApiType(api_type) if isinstance(api_type, str) else api_type
        
        if api_version is None: 
            self.api_version = self.api_type.get_version(api_version)
        else:
            self.api_version = api_version
        if organization is not None: self.organization = organization
        if proxies is not None: self.proxies = proxies
        if timeout is not None: self.timeout = timeout
        if max_retries is not None: self.max_retries = max_retries
        if app_info is not None: self.app_info = app_info
        if debug_enabled is not None: self.debug_enabled = debug_enabled
        if ignore_errors is not None: self.ignore_errors = ignore_errors
        if disable_retries is not None: self.disable_retries = disable_retries

        if max_connections is not None: self.max_connections = max_connections
        if max_keepalive_connections is not None: self.max_keepalive_connections = max_keepalive_connections
        if keepalive_expiry is not None: self.keepalive_expiry = keepalive_expiry
        if custom_headers is not None: self.ctx.update_headers(custom_headers)

    def disable_httpx_logger(self):
        """
        Disables the HTTPX Logger
        """
        from ..utils.logs import configure_httpx_logger
        configure_httpx_logger(logging.ERROR)
    
    def enable_httpx_logger(self):
        """
        Enables the HTTPX Logger
        """
        from ..utils.logs import configure_httpx_logger
        configure_httpx_logger(logging.INFO)


    @eproperty
    def proxy(self) -> OpenAIProxySettings:
        """
        Return the Proxy Settings
        """
        return OpenAIProxySettings()


from __future__ import annotations

from pydantic.alias_generators import to_camel
from typing import Any, Dict, Type, Union


def dict_diff(dict_a: Dict[str, Any], dict_b: Dict[str, Any], show_value_diff: bool = True):
    
    """
    Compare two dictionaries and return the difference between them
    """
    result = {
        'added': {k: dict_b[k] for k in set(dict_b) - set(dict_a)},
        'removed': {k: dict_a[k] for k in set(dict_a) - set(dict_b)},
    }
    if show_value_diff:
        common_keys =  set(dict_a) & set(dict_b)
        result['value_diffs'] = {
            k:(dict_a[k], dict_b[k])
            for k in common_keys
            if dict_a[k] != dict_b[k]
        }
    return result


def format_table_name(table_name: str) -> str:
    """
    Formats the table name

    >>> format_table_name('my.test.table_name') 
    'my_test_tableName'
    """
    parts = table_name.split('.')
    parts = [to_camel(part) for part in parts]
    return '_'.join(parts)
